/***************************************************

    mcmr.c : core of MCMR queue simulator

****************************************************/
#include <stdio.h>
#include <assert.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <stdlib.h>
#include <math.h>

#include "sys_mcmr.h"
#include "mcmr.h"

/* globals */
resource_type R, R2;
class_type C, C2;
std_count_type **CNT, **CNT2; /* holds current measure */
control_type CTR;
event_list_type E;
callback_type CB;
checkpoint_struc_type *CS = NULL;
int nNumCS;

double *Lambdas;
double dLambda;

extern double rint(double);

int ( *ArrRateFunction )( double, int, double* );
int ( *SerRateFunction )( double, int, int, double* );
int ( *DumpFunction )( double, int, double**, double**, double, void* );
void *DumpParam;

/* function declarations */

void Evaluate();
void EvaluateSingle();
int IsBlocked(int);

void AllocCountMeasure(double, std_count_type ***);
void FreeCountMeasure(double, std_count_type **);
void ClearCountMeasure();
void CountMeasure(int, int, int, int, double);
void AvgCountMeasure();
void FreeC(class_type *cp);
void FreeR(resource_type *rp);
void MirrorRandC(resource_type *, resource_type *,
                 class_type *, class_type *);

void BuildLambdaList();
int Double_Equal(double, double);

void Set_ClassID();

int cid2idx(int);
int rid2idx(int);
int rid2cidx(int, int);
int idx2cid(int);
int idx2rid(int);
int cid2idx_at(double, int);
int rid2idx_at(double, int);
int rid2cidx_at(double, int, int);
int idx2cid_at(double, int);

void Check_CS();
int Get_CS(double);
void Print_CS();

void DumpAvg();

#ifdef DEBUG
extern int getrusage(int who, struct rusage *rusage);
#endif

/* repeat N times as given by the user */
void Evaluate()
{
    int j;
    struct timeval tv;
    struct timezone tzp;

    /* init seed */
    gettimeofday( &tv, &tzp );
    srand48( tv.tv_usec );

    /* set classID's for R */
    Set_ClassId();

    /* get the mirror of C structure */
    Check_CS();

#ifdef DEBUG
{
struct rusage ru;
getrusage(RUSAGE_SELF, &ru);
fprintf(stderr, "In Evaluate() Before AllocCountmeasure() %d, %d, %d\n",
ru.ru_maxrss, ru.ru_idrss, ru.ru_nswap);
fflush(stderr);
}
#endif

    /* allocate memory space for standard measures */
    AllocCountMeasure(0.0, &CNT);
    AllocCountMeasure(0.0, &CNT2);

#ifdef DEBUG
{
struct rusage ru;
getrusage(RUSAGE_SELF, &ru);
fprintf(stderr, "In Evaluate() After AllocCountmeasure() %d, %d, %d\n",
ru.ru_maxrss, ru.ru_idrss, ru.ru_nswap);
fflush(stderr);
}
#endif

    /* Store the mirror image of R and C */
    MirrorRandC(&R, &R2, &C, &C2); 

    for (CTR.nCurrentRun = 0; CTR.nCurrentRun < CTR.nSimulationRuns ;
         CTR.nCurrentRun ++ ) {

#ifdef DEBUG
fprintf(stderr, "In Evaluate() run = %d\n", CTR.nCurrentRun);
fflush(stderr);
#endif

        /* reset class and resource */ 
        FreeR(&R);
        FreeC(&C);
        MirrorRandC(&R2, &R, &C2, &C); 

        /* build the arrival rate list */
        BuildLambdaList();

        /* initialize control parameters */
        CTR.dCurrentTime = 0.0;
        CTR.nCurrentEventRecord = 0;
        CTR.nCurrentMeasureRecord = 0;

        /* clear count variables */
        ClearCountMeasure();

		/* reset event list */
		ResetEventlistHdr();

        /* call the main loop */ 
        EvaluateSingle();

#ifdef DEBUG
fprintf(stderr, "In Evaluate() After EvaluateSingle()\n"); 
fflush(stderr);
#endif

        /* average measures */
        AvgCountMeasure();

        /* if last run, call dump function */

#ifdef DEBUG
fprintf(stderr, "In Evaluate() After AvgCountMeasure()\n"); 
fflush(stderr);
#endif
        if (CTR.nCurrentRun == CTR.nSimulationRuns - 1) {
            DumpAvg();
        }

#ifdef DEBUG
{
struct rusage ru;
getrusage(RUSAGE_SELF, &ru);
fprintf(stderr, "In Evaluate() After AvgCountMeasure() %d, %d, %d\n",
ru.ru_maxrss, ru.ru_idrss, ru.ru_nswap);
fflush(stderr);
}
#endif


    }
}
 
/* main loop of the simulator */
void EvaluateSingle()
{
    int  j, adjusts;
    resource_usage_type *ru;
    double dLastDump = 0.0, dSave;
	event_type *ev;

    while (CTR.dCurrentTime <= CTR.dSimulationTime) {

        /* 1. if CTR.dCurrentTime == 0, generate arrivals
             for all the classes */ 
        if (CTR.dCurrentTime == 0.0) {
            AddEvent(ARRIVAL, 0);
        }

        /* 2. handle different types of events */ 
        ev = FetchNextEvent(); 
        if (ev == NULL) break;

#ifdef DEBUG
fprintf(stderr, "Time = %f\n", CTR.dCurrentTime);
fprintf(stderr, " Type = %d, Class = %d, \n", eventtype(ev), 
eventclass(ev)); 
fflush(stderr);
#endif

        /* Note that event ev is the 'current' event */
        switch (eventtype(ev)) {

            case ARRIVAL :
                /* check if it is blocked */
                if (IsBlocked(eventclass(ev)) == FALSE) {

                    /* allocate the amount required for each resource */
                    ru = C.sResources[ eventclass(ev) ];
                    for (j=0;j < C.nNumResources[eventclass(ev)];j++) {
                        R.dAvailable[ru[j].nIndex] -= ru[j].dRequire;
                    }

                    /* set up departing time and get next arrival */
                    AddEvent(DEPARTURE, eventclass(ev));

                    /* increase the counter */
                    if ( C.nNumResources[eventclass(ev)] ) {
                    
                        /* if no resource then just pass through above lines */
                        CountMeasure(COUNT_ACCEPTED_CLASS, 
                            eventclass(ev), 1, 0, CTR.dCurrentTime);
                    }
                } else {
                    CountMeasure(COUNT_BLOCKED_CLASS, eventclass(ev), 1, 0, CTR.dCurrentTime);
                }

                /* generate the next arrival */
                AddEvent(ARRIVAL, 0); 

                /* delete from the event list */
                DeleteEvent(ev);

                break;

            case DEPARTURE :    
                /* NOTE : Assume every customer relinquishes its 
                          resources at once 
                */

                /* deallocate the amount required for each resource */
                ru = C.sResources[eventclass(ev)];
                for (j=0;j < C.nNumResources[eventclass(ev)];j++) {
                    R.dAvailable[ru[j].nIndex] += ru[j].dRequire;
                }

                /* delete from the event list */
                DeleteEvent(ev);

                /* increase the counter */
                if ( C.nNumResources[eventclass(ev)] ) {
                    CountMeasure(COUNT_DEPART, eventclass(ev), 
                                    1, 0, CTR.dCurrentTime);
                }

                break;

            case CALLBACK: 
                CB.callbacks[((cb_event_type *)eventclass(ev))->nId].f(
                    CTR.dCurrentTime, ((cb_event_type *)eventclass(ev))->nParam,
                    ((cb_event_type *)eventclass(ev))->pParam);
				DeleteEvent(ev);

                break;

            case CALLBACKINIT:
                CB.callbacks[((cb_event_type *)eventclass(ev))->nId].f(
                    CTR.dCurrentTime, ((cb_event_type *)eventclass(ev))->nParam,
                    ((cb_event_type *)eventclass(ev))->pParam);
				AdvanceEventlistHdr(ev);

                break;

            default :
                assert(1); 
                break;
        }
    }
}

/* return TRUE if the customer in class 'ClassId' can't allocate
  any of all the resources it's supposed to get
*/ 
int IsBlocked(int ClassId)
{
    int i, j, flag = FALSE;
    resource_usage_type *ru;

    ru = C.sResources[ClassId];
    for (i=0;i < C.nNumResources[ClassId];i++) {
        if (R.dAvailable[C.sResources[ClassId][i].nIndex] < 
                C.sResources[ClassId][i].dRequire) {
            flag = TRUE;
            CountMeasure(COUNT_BLOCKED, ClassId, 1, i, CTR.dCurrentTime); 
        } else {
            CountMeasure(COUNT_PASSED, ClassId, 1, i, CTR.dCurrentTime); 
        } 
    }

    return flag;
}

void AllocCountMeasure(double time, std_count_type ***cp)
{
    int i, j, k, numdumps, startidx;

    /* 1. Initialize (*cp) */
    numdumps = CTR.dSimulationTime / CTR.dCollectStdMeasure + 2;
    assert(numdumps);

    if ( (*cp) == NULL ) { 
        (*cp) = (std_count_type **)malloc(numdumps * 
            sizeof(std_count_type *));
        assert((*cp));
    }

    startidx = rint(time / CTR.dCollectStdMeasure);

#ifdef DEBUG
fprintf(stderr, "In AllocCountMeasure(%f), idx = %d\n", time, startidx);
fflush(stderr);
#endif

  for (j=startidx;j < numdumps;j++) {

    (*cp)[j] = (std_count_type *)malloc(
                                sizeof(std_count_type));
    assert((*cp)[j]);

    k = Get_CS(time);


    /* dDepart_c_r, dBlocked_c_r, dAccepted_c_r, dPassed_c_r */
    (*cp)[j]->dDepart_c_r = (double **)malloc(CS[k].nCSetSize * 
												sizeof(double));
    assert((*cp)[j]->dDepart_c_r);
    (*cp)[j]->dBlocked_c_r = (double **)malloc(CS[k].nCSetSize * 
												sizeof(double));
    assert((*cp)[j]->dBlocked_c_r);
    (*cp)[j]->dAccepted_c_r = (double **)malloc(CS[k].nCSetSize * 
												sizeof(double));
    assert((*cp)[j]->dAccepted_c_r);
    (*cp)[j]->dPassed_c_r = (double **)malloc(CS[k].nCSetSize * 
												sizeof(double));
    assert((*cp)[j]->dPassed_c_r);

    for (i=0;i < CS[k].nCSetSize;i++) {
        (*cp)[j]->dDepart_c_r[i] = (double *)malloc(CS[k].nNumResources[i] * 
                              sizeof(double));
        assert((*cp)[j]->dDepart_c_r[i]);
        memset((*cp)[j]->dDepart_c_r[i], 0, CS[k].nNumResources[i] * 
							sizeof(double));
        (*cp)[j]->dBlocked_c_r[i] = (double *)malloc(CS[k].nNumResources[i] * 
                              sizeof(double));
        assert((*cp)[j]->dBlocked_c_r[i]);
        memset((*cp)[j]->dBlocked_c_r[i], 0, CS[k].nNumResources[i] * 
							sizeof(double));
        (*cp)[j]->dAccepted_c_r[i] = (double *)malloc(CS[k].nNumResources[i] * 
                              sizeof(double));
        assert((*cp)[j]->dAccepted_c_r[i]);
        memset((*cp)[j]->dAccepted_c_r[i], 0, CS[k].nNumResources[i] * 
							  sizeof(double));
        (*cp)[j]->dPassed_c_r[i] = (double *)malloc(CS[k].nNumResources[i] * 
                              sizeof(double));
        assert((*cp)[j]->dPassed_c_r[i]);
        memset((*cp)[j]->dPassed_c_r[i], 0, CS[k].nNumResources[i] * 
							  sizeof(double));
    }

    /* dAccepted_c */
    (*cp)[j]->dAccepted_c = (double *)malloc(CS[k].nCSetSize * sizeof(double));
    assert((*cp)[j]->dAccepted_c);
    memset((*cp)[j]->dAccepted_c, 0, CS[k].nCSetSize * sizeof(double));
    /* dDepart_c */
    (*cp)[j]->dDepart_c = (double *)malloc(CS[k].nCSetSize * sizeof(double));
    assert((*cp)[j]->dDepart_c);
    memset((*cp)[j]->dDepart_c, 0, CS[k].nCSetSize * sizeof(double));
    /* dBlocked_c */
    (*cp)[j]->dBlocked_c = (double *)malloc(CS[k].nCSetSize * sizeof(double));
    assert((*cp)[j]->dBlocked_c);
    memset((*cp)[j]->dBlocked_c, 0, CS[k].nCSetSize * sizeof(double));
  }

#ifdef DEBUG
fprintf(stderr, "End AllocCountMeasure(%f), idx = %d\n", time, startidx);
fflush(stderr);
#endif

}

void ClearCountMeasure()
{
    int i, j, k, numdumps;

    numdumps = CTR.dSimulationTime / CTR.dCollectStdMeasure + 2;

    /* clear count variables */
    for (j=0;j < numdumps;j++) {
      k = Get_CS(j * CTR.dCollectStdMeasure);

#ifdef DEBUG
fprintf(stderr, "time = %f, CS[k=%d], ClearCountMeasure() = %d\n", 
j * CTR.dCollectStdMeasure, k, j);
fflush(stderr);
#endif

      for (i=0;i < CS[k].nCSetSize;i++) {
        memset(CNT2[j]->dDepart_c_r[i], 0, 
               CS[k].nNumResources[i] * sizeof(double));
        memset(CNT2[j]->dBlocked_c_r[i], 0, 
               CS[k].nNumResources[i] * sizeof(double));
        memset(CNT2[j]->dAccepted_c_r[i], 0, 
               CS[k].nNumResources[i] * sizeof(double));
        memset(CNT2[j]->dPassed_c_r[i], 0, 
               CS[k].nNumResources[i] * sizeof(double));
      }
      memset(CNT2[j]->dAccepted_c, 0, CS[k].nCSetSize * sizeof(double));
      memset(CNT2[j]->dBlocked_c, 0, CS[k].nCSetSize * sizeof(double));
      memset(CNT2[j]->dDepart_c, 0, CS[k].nCSetSize * sizeof(double));
    }
}

void FreeCountMeasure(double time, std_count_type **cp)
{
    int i, j, k, numdumps, startidx;

    startidx = rint(time / CTR.dCollectStdMeasure);
    numdumps = CTR.dSimulationTime / CTR.dCollectStdMeasure + 2;

#ifdef DEBUG
fprintf(stderr, "In FreeCountMeasure(%f), idx = %d\n", time, startidx);
fflush(stderr);
#endif

   /* Free cp */
   for (j=startidx;j < numdumps;j++) {
     k = Get_CS(time);      
     for (i=0;i < CS[k].nCSetSize;i++) {
        free(cp[j]->dDepart_c_r[i]);
        free(cp[j]->dBlocked_c_r[i]);
        free(cp[j]->dAccepted_c_r[i]);
        free(cp[j]->dPassed_c_r[i]);
     }
     free(cp[j]->dAccepted_c);
     free(cp[j]->dBlocked_c);
   }

#ifdef DEBUG
fprintf(stderr, "End FreeCountMeasure(%f), idx = %d\n", time, startidx);
fflush(stderr);
#endif

}

void CountMeasure(int Which, int ClassId, int Delta, int ResourceId, double time)
{
    int i, idx;
    
    idx = rint(time / CTR.dCollectStdMeasure);

if ( ClassId >= C.nSetSize ||
     ResourceId >= C.nNumResources[ClassId] ) {
#ifdef DEBUG
fprintf(stderr, "Got it! %d\n", CTR.nCurrentRun); 
fflush(stderr);
#endif
return;
}

    switch(Which) {
/*
        case COUNT_ACCEPTED:
            CNT2[idx]->dAccepted_c_r[ClassId][ResourceId] += Delta;
            break;
*/

        case COUNT_ACCEPTED_CLASS:
            for (i=0;i < C.nNumResources[ClassId];i++) {
                CNT2[idx]->dAccepted_c_r[ClassId][i] += Delta;  
            }
            CNT2[idx]->dAccepted_c[ClassId] += Delta;

            break;

        case COUNT_DEPART:
            for (i=0;i < C.nNumResources[ClassId];i++) {
                CNT2[idx]->dDepart_c_r[ClassId][i] += Delta;  
            }

            CNT2[idx]->dDepart_c[ClassId] += Delta;

            break;

        case COUNT_BLOCKED:
            CNT2[idx]->dBlocked_c_r[ClassId][ResourceId] += Delta;  
            break;

        case COUNT_BLOCKED_CLASS:
            CNT2[idx]->dBlocked_c[ClassId] += Delta;
            break;

        case COUNT_PASSED:
            CNT2[idx]->dPassed_c_r[ClassId][ResourceId] += Delta;
            break;

        default : assert(0);
    }
}

void AvgCountMeasure()
{
    int i, j, idx, ClassIdx, ResourceIdx;
    double Delta;
    
    for (idx=0;idx <= CTR.dSimulationTime / CTR.dCollectStdMeasure;idx++) {

        /* find the appropriate CS structure */
        j = Get_CS( idx * CTR.dCollectStdMeasure);

#ifdef DEBUG
fprintf(stderr, "AvgCountMeasure() time = %f, idx = %d, CS[%d] \n", 
idx * CTR.dCollectStdMeasure, idx, j);
fflush(stderr);
#endif

        for (ClassIdx = 0; ClassIdx < CS[j].nCSetSize;ClassIdx++) {
         for (ResourceIdx = 0;ResourceIdx < CS[j].nNumResources[ClassIdx];
                                                        ResourceIdx++) {

#ifdef DEBUG
fprintf(stderr, "AvgCountMeasure() within 1st loop cid:%d rid:%d\n",
ClassIdx, ResourceIdx); 
fflush(stderr);
#endif

            Delta = CNT2[idx]->dAccepted_c_r[ClassIdx][ResourceIdx];
            CNT[idx]->dAccepted_c_r[ClassIdx][ResourceIdx] = 
              ( (double)CTR.nCurrentRun * 
                CNT[idx]->dAccepted_c_r[ClassIdx][ResourceIdx] + Delta ) 
            / ( CTR.nCurrentRun + 1);

            Delta = CNT2[idx]->dDepart_c_r[ClassIdx][ResourceIdx];
            CNT[idx]->dDepart_c_r[ClassIdx][ResourceIdx] = 
              ( (double)CTR.nCurrentRun * 
                CNT[idx]->dDepart_c_r[ClassIdx][ResourceIdx] + Delta ) 
             / ( CTR.nCurrentRun + 1);

            Delta = CNT2[idx]->dBlocked_c_r[ClassIdx][ResourceIdx];
            CNT[idx]->dBlocked_c_r[ClassIdx][ResourceIdx] = 
              ( (double)CTR.nCurrentRun * 
                CNT[idx]->dBlocked_c_r[ClassIdx][ResourceIdx] + Delta ) 
            / ( CTR.nCurrentRun + 1);

            Delta = CNT2[idx]->dPassed_c_r[ClassIdx][ResourceIdx];
            CNT[idx]->dPassed_c_r[ClassIdx][ResourceIdx] = 
              ( (double)CTR.nCurrentRun * 
                CNT[idx]->dPassed_c_r[ClassIdx][ResourceIdx] + Delta ) 
            / ( CTR.nCurrentRun + 1);
         }
        }

#ifdef DEBUG
fprintf(stderr, "AvgCountMeasure() after 1st loop\n"); 
fflush(stderr);
#endif

        for (ClassIdx = 0; ClassIdx < CS[j].nCSetSize;ClassIdx++) {
            Delta = CNT2[idx]->dAccepted_c[ClassIdx];
            CNT[idx]->dAccepted_c[ClassIdx] = 
              ( (double)CTR.nCurrentRun * 
                CNT[idx]->dAccepted_c[ClassIdx] + Delta ) 
            / ( CTR.nCurrentRun + 1);

            Delta = CNT2[idx]->dDepart_c[ClassIdx];
            CNT[idx]->dDepart_c[ClassIdx] = 
              ( (double)CTR.nCurrentRun * 
                CNT[idx]->dDepart_c[ClassIdx] + Delta ) 
            / ( CTR.nCurrentRun + 1);

            Delta = CNT2[idx]->dBlocked_c[ClassIdx];
            CNT[idx]->dBlocked_c[ClassIdx] = 
              ( (double)CTR.nCurrentRun * 
                CNT[idx]->dBlocked_c[ClassIdx] + Delta ) 
            / ( CTR.nCurrentRun + 1);
        }
    }
}

int Double_Equal(double one, double two)
{
    #define THRESHOLD 0.00000001

    if (fabs(one - two) <= THRESHOLD) {
        return 1;
    }
    else {
        return 0;
    }
}

void BuildLambdaList()
{
    int j, k;
    double r = 0.0;

    /* allocate space for lambda list */
    Lambdas = (double *)realloc( Lambdas, C.nSetSize * sizeof(double) );
    assert(Lambdas);

    dLambda = 0.0;
    for (j=0;j < C.nSetSize;j++) {
        r += C.dArrRate[j];
        Lambdas[j] = r;
        dLambda += C.dArrRate[j];
    }
}

void FreeC(class_type *cp)
{
    int i;

    free(cp->id);
    free(cp->inuse);
    free(cp->dArrRate);
    free(cp->nNumResources);
    for (i=0;i < cp->nSetSize;i++) {
        free(cp->sResources[i]);
    }
    free(cp->sResources);
}

void FreeR(resource_type *rp)
{
    int i;

    free(rp->id);
    free(rp->inuse);
    free(rp->dCapacity);
    free(rp->dAvailable);
    free(rp->nNumClasses);
    for (i=0;i < rp->nSetSize;i++) {
        free(rp->nClassId[i]);
    }
    free(rp->nClassId);
}

void MirrorRandC(resource_type *rs, resource_type *rd,
                 class_type *cs, class_type *cd)
{
    int i;

    /* resource set rs -> rd */
    rd->nSetSize = rs->nSetSize;
    rd->id = (int *)malloc(rd->nSetSize * sizeof(int));
    rd->inuse = (int *)malloc(rd->nSetSize * sizeof(int));
    rd->nNumClasses = (int *)malloc(rd->nSetSize * sizeof(int));
    rd->dCapacity = (double *)malloc(rd->nSetSize * sizeof(double));
    rd->dAvailable = (double *)malloc(rd->nSetSize * sizeof(double));
    assert ( rd->id && rd->inuse && rd->nNumClasses && rd->dCapacity &&
             rd->dAvailable );

    memcpy(rd->id, rs->id, rd->nSetSize * sizeof(int)); 
    memcpy(rd->inuse, rs->inuse, rd->nSetSize * sizeof(int)); 
    memcpy(rd->nNumClasses, rs->nNumClasses, rd->nSetSize * sizeof(int)); 
    memcpy(rd->dCapacity, rs->dCapacity, rd->nSetSize * sizeof(double)); 
    memcpy(rd->dAvailable, rs->dAvailable, rd->nSetSize * sizeof(double)); 

    rd->nClassId = (int **)malloc(rd->nSetSize * sizeof(int *));
    assert(rd->nClassId);
    for (i=0;i < rd->nSetSize;i++) {
        rd->nClassId[i] = malloc( rd->nNumClasses[i] * sizeof(int) );
        assert( rd->nClassId[i] );
        memcpy( rd->nClassId[i], rs->nClassId[i], 
                rd->nNumClasses[i] * sizeof(int) );
    }

    /* class set cs -> cd */
    cd->nSetSize = cs->nSetSize;
    cd->id = (int *)malloc(cd->nSetSize * sizeof(int));
    cd->inuse = (int *)malloc(cd->nSetSize * sizeof(int));
    cd->dArrRate = (double *)malloc(cd->nSetSize * sizeof(double)); 
    cd->nNumResources = (int *)malloc(cd->nSetSize * sizeof(int));
    assert( cd->id && cd->inuse && cd->dArrRate && cd->nNumResources );

    memcpy(cd->id, cs->id, cd->nSetSize * sizeof(int)); 
    memcpy(cd->inuse, cs->inuse, cd->nSetSize * sizeof(int)); 
    memcpy(cd->nNumResources, cs->nNumResources, cd->nSetSize * sizeof(int)); 
    memcpy(cd->dArrRate, cs->dArrRate, cd->nSetSize * sizeof(double)); 

    cd->sResources = (resource_usage_type**)malloc( cd->nSetSize *
                                            sizeof(resource_usage_type*) );
    assert(cd->sResources);
    for (i=0;i < cd->nSetSize;i++) {
        cd->sResources[i] = (resource_usage_type*)
                            malloc( cd->nNumResources[i] * 
                                    sizeof(resource_usage_type) );
        assert( cd->sResources[i] );
        memcpy( cd->sResources[i], cs->sResources[i], 
                cd->nNumResources[i] * sizeof(resource_usage_type) );
    }
}

void Set_ClassId()
{
    int i, j, *cnt;

    /* make sure we have classes defined */
    assert(C.nSetSize);

    /* allocate cnt[] */
    cnt = (int *)malloc(R.nSetSize * sizeof(int));
    assert(cnt);
    memset(cnt, 0, R.nSetSize * sizeof(int));

    /* count */
    for (i=0;i < C.nSetSize;i++) {
        for (j=0;j < C.nNumResources[i];j++) {
            cnt[ C.sResources[i][j].nIndex ] ++;
        }
    }

    /* allocate spaces */
    R.nNumClasses = (int *)malloc(R.nSetSize * sizeof(int));
    assert(R.nNumClasses);
    R.nClassId = (int **)malloc(R.nSetSize * sizeof(int *));
    assert(R.nClassId);
    for (i=0;i < R.nSetSize;i++) {
        R.nClassId[i] = (int *)malloc( cnt[i] * sizeof(int) );
        assert(R.nClassId[i]);
    }

    /* fill up the structure */
    memcpy(R.nNumClasses, cnt, R.nSetSize * sizeof(int));
    memset(cnt, 0, R.nSetSize * sizeof(int));
    for (i=0;i < C.nSetSize;i++) {
        for (j=0;j < C.nNumResources[i];j++) {
            R.nClassId[ C.sResources[i][j].nIndex ]
                      [ cnt[ C.sResources[i][j].nIndex ]++ ] = i;
        }
    }

    free(cnt);
}

int cid2idx(int cid)
{
    int i=0;

    while(i < C.nSetSize && cid != C.id[i]) i++;
    if (i == C.nSetSize) {
        return -1;
    } else {
        return i;
    }
}

int rid2idx(int rid)
{
    int i=0;

    while(i < R.nSetSize && rid != R.id[i]) i++;
    if (i == R.nSetSize) {
        return -1;
    } else {
        return i;
    }
}

int rid2cidx(int cid, int rid)
{
    int i=0;
    int cidx, ridx;

    cidx = cid2idx(cid);
    if (cidx < ERR_SUCC) {
        return ERR_INVALID_CLASSID;
    }

    ridx = rid2idx(rid);

    while(i < C.nNumResources[cidx] && 
          ridx != C.sResources[cidx][i].nIndex) i++;
    if (i == C.nNumResources[cidx]) {
        return -1;
    } else {
        return i;
    }
}

int idx2cid(int idx)
{
    if (idx >= C.nSetSize) {
        return -1;
    } else {
        return C.id[idx];
    }
}



int idx2rid(int idx)
{
    if (idx >= R.nSetSize) {
        return -1;
    } else {
        return R.id[idx];
    }
}

int idx2cid_at(double time, int idx)
{
    int cs;

    cs = Get_CS(time);
    if (idx >= CS[cs].nCSetSize) {
        return -1;
    } else {
        return CS[cs].cid[idx];
    }
}

int idx2rid_at(double time, int idx)
{
    int cs;

    cs = Get_CS(time);
    if (idx >= CS[cs].nRSetSize) {
        return -1;
    } else {
        return CS[cs].rid[idx];
    }
}

int cid2idx_at(double time, int cid)
{
    int i=0, cs;

    cs = Get_CS(time);

    while(i < CS[cs].nCSetSize && cid != CS[cs].cid[i]) i++;
    if (i == CS[cs].nCSetSize) {
        return -1;
    } else {
        return i;
    }
}

int rid2idx_at(double time, int rid)
{
    int i=0, rs;

    while(i < CS[rs].nRSetSize && rid != CS[rs].rid[i]) i++;
    if (i == CS[rs].nRSetSize) {
        return -1;
    } else {
        return i;
    }
}

int rid2cidx_at(double time, int cid, int rid)
{
    int i=0;
    int cidx, ridx, cs;

    cidx = cid2idx_at(time, cid);
    if (cidx < ERR_SUCC) {
        return ERR_INVALID_CLASSID;
    }

    ridx = rid2idx_at(time, rid);

    cs = Get_CS(time);

    while(i < CS[cs].nNumResources[cidx] && 
          ridx != CS[cs].sResources[cidx][i].nIndex) i++;
    if (i == CS[cs].nNumResources[cidx]) {
        return -1;
    } else {
        return i;
    }
}

/***************************************************
    Functions to manage CS
****************************************************/
void Check_CS()
{
    int i;

    if ( (nNumCS == 0) || 
         (nNumCS && 
         CS[ nNumCS-1 ].timestamp !=  
         CTR.dCurrentTime / CTR.dCollectStdMeasure ) ) {

        nNumCS++;

        CS = (checkpoint_struc_type *)realloc( CS, nNumCS * 
                              sizeof(checkpoint_struc_type) );  
        assert(CS);
    }

    CS[ nNumCS-1 ].timestamp = rint(CTR.dCurrentTime / CTR.dCollectStdMeasure);
    CS[ nNumCS-1 ].nCSetSize = C.nSetSize;
    CS[ nNumCS-1 ].nRSetSize = R.nSetSize;
    CS[ nNumCS-1 ].cid = (int *)malloc( C.nSetSize * sizeof(int) );
    assert( CS[ nNumCS-1 ].cid );
    memcpy( CS[ nNumCS-1 ].cid, C.id, C.nSetSize * sizeof(int));

    CS[ nNumCS-1 ].rid = (int *)malloc( R.nSetSize * sizeof(int) );
    assert( CS[ nNumCS-1 ].rid );
    memcpy( CS[ nNumCS-1 ].rid, R.id, R.nSetSize * sizeof(int));

    CS[ nNumCS-1 ].dCapacity = (double *)malloc( R.nSetSize * sizeof(double) );
    assert( CS[ nNumCS-1 ].dCapacity );
    memcpy( CS[ nNumCS-1 ].dCapacity, R.dCapacity, R.nSetSize * sizeof(double));

    CS[ nNumCS-1 ].nNumResources = (int *)malloc( C.nSetSize *
                                            sizeof( int )     );
    assert( CS[ nNumCS-1 ].nNumResources );
    memcpy( CS[ nNumCS-1 ].nNumResources, C.nNumResources, 
            C.nSetSize * sizeof( int )                    );

    CS[ nNumCS-1 ].sResources = (resource_usage_type **)malloc(
                        C.nSetSize * sizeof(resource_usage_type *));
    assert(CS[ nNumCS-1 ].sResources);
            
    for (i=0;i < C.nSetSize;i++) {
        CS[ nNumCS-1 ].sResources[i] =
            (resource_usage_type *)malloc( C.nNumResources[i] * sizeof(
                                           resource_usage_type) );
        assert( CS[ nNumCS-1 ].sResources[i] );
        memcpy( CS[ nNumCS-1 ].sResources[i], C.sResources[i],
                C.nNumResources[i] * sizeof(resource_usage_type) );
    }

#ifdef DEBUG
fprintf(stderr, "in Check_CS(): idx = %d, time = %f\n", 
                nNumCS-1, CTR.dCurrentTime             );
fflush(stderr);
Print_CS(nNumCS-1);
#endif
}

int Get_CS(double time)
{
    int i=0;

    while ( i < nNumCS && 
            CS[i].timestamp <= rint(time / CTR.dCollectStdMeasure) ) i++;

#ifdef DEBUG
if (CTR.nCurrentRun == 12) {
fprintf(stderr, "in Get_CS(%f), idx=%d\n", time, i-1);
fprintf(stderr, "in CS[i].timestamp %d, time / CTR.dCollectStdMeasure=%f\n",
CS[i-1].timestamp, time / CTR.dCollectStdMeasure); 
fflush(stderr);
}
#endif

    return i-1;
}

void Print_CS(int idx)
{
    int i, j;

    printf("-----------------CS(%d) : timestamp = %d\n", idx, 
            CS[idx].timestamp);
    printf(" --------- Classes : CS(%d) -------------- \n", idx);
    for (i=0;i < CS[idx].nCSetSize;i++) {
        printf("%3d : NumResource = %3d\n",
                CS[idx].cid[i], CS[idx].nNumResources[i]);
        for (j=0;j < CS[idx].nNumResources[i];j++) {
            printf("\t resource %3d : Require = %6.2f, SerRate = %6.2f\n",
                    C.sResources[i][j].nIndex, C.sResources[i][j].dRequire,
                    C.sResources[i][j].dSerRate);
        }
    }
    printf(" --------- Resources : CS(%d) -------------- \n", idx);
    for (i=0;i < CS[idx].nRSetSize;i++) {
        printf("%3d ", CS[idx].rid[i]);
    }
    printf("\n\n");
}

void DumpAvg()
{
    int i, j, k, cs, cid, rid, midx;
    double *dN, *dBp, dTotalAdmission;
    typedef struct ridN_tag {
        int nRid;
        double dN;
    } ridN_type; 
    typedef struct cidN_tag {
        int nCid;
        int nNum;
        ridN_type *sN;
    } cidN_type; 
    cidN_type **sCidN = 0;
    int nsCidNSize, nNcid; 
    cidN_type **tmp1;
    cidN_type *tmp2;
    ridN_type *tmp3;

    for (midx=0, CTR.dCurrentTime = 0.0;CTR.dCurrentTime < CTR.dSimulationTime; 
         CTR.dCurrentTime += CTR.dCollectStdMeasure, midx++) {
        
        /* locate check point */
        cs = Get_CS(CTR.dCurrentTime);

        /* allocate sCidN[] */
        tmp1 = 
            (cidN_type**)malloc(CS[cs].nCSetSize * sizeof(cidN_type*));
        assert(tmp1);
        memset(tmp1, 0, CS[cs].nCSetSize * sizeof(cidN_type*));
        if (sCidN) { /* if not first time */
            memcpy(tmp1, sCidN, nsCidNSize * sizeof(cidN_type*));
            free(sCidN);
        }
        sCidN = tmp1;
        nsCidNSize = CS[cs].nCSetSize;

        for (i=0;i < CS[cs].nCSetSize;i++) { /* for each class */

            /* determine cid */
            cid = idx2cid_at(CTR.dCurrentTime, i);

            /* determine nNcid */
            for (nNcid=0;nNcid < nsCidNSize;nNcid++) {
                if ( sCidN[nNcid] && cid == sCidN[nNcid]->nCid )
                    break; 
            }
            if (nNcid == nsCidNSize) { /* first time */
                /* first unused entry */
                for (j=0;j < nsCidNSize;j++) {
                    if (sCidN[j] == NULL)
                        break;
                }
                assert(j < nsCidNSize);
                nNcid = j;
            }


            if (cid > -1) { /* if the class exists */

                /* ----------------------------------------------- */
                /* calculate N[]                                   */
                /* ----------------------------------------------- */
                tmp2 = (cidN_type *)malloc(sizeof(cidN_type));
                tmp3 = (ridN_type *)malloc(CS[cs].nNumResources[i] *
                                            sizeof(ridN_type)           );
                assert(tmp2 && tmp3);
                memset(tmp3, 0, CS[cs].nNumResources[i] * sizeof(ridN_type)); 
                for (j=0;j < CS[cs].nNumResources[i];j++) {
                    tmp3[j].nRid = -1;
                    tmp3[j].dN = 0.0;
                }

                if (sCidN[nNcid]) {
                    memcpy(tmp2, sCidN[nNcid], sizeof(cidN_type)); 
                    memcpy(tmp3, sCidN[nNcid]->sN, 
                                 sCidN[nNcid]->nNum * sizeof(ridN_type)); 
                    free(sCidN[nNcid]->sN);
                    free(sCidN[nNcid]);
                }
                tmp2->nNum = CS[cs].nNumResources[i];
                tmp2->nCid = cid; 
                tmp2->sN = tmp3; 
                sCidN[nNcid] = tmp2; 

#ifdef DEBUG
fprintf(stderr, " In DumpAvg() i=%d, cid=%d, numRes=%d\n", i, cid,
CS[cs].nNumResources[i]); 
fflush(stderr);
#endif

                /* for each resource */
                for (j=0;j < CS[cs].nNumResources[i];j++) { 

                    /* get rid */
                    rid = idx2rid_at( CTR.dCurrentTime, 
                                      CS[cs].sResources[i][j].nIndex );

                    /* fetch the right entry */ 
                    for (k=0;k < CS[cs].nNumResources[i];k++) {
                        if ( rid == sCidN[nNcid]->sN[k].nRid ) {
                            break;
                        }
                    }
                    if (k == CS[cs].nNumResources[i]) { /* first time */
                        for (k=0;k < CS[cs].nNumResources[i];k++) {
                            if ( sCidN[nNcid]->sN[k].nRid == -1 )
                                break;
                        }
                    }
					assert( k < CS[cs].nNumResources[i] );

                    /* set the value */
                    sCidN[nNcid]->sN[k].dN += CNT[midx]->dAccepted_c_r[i][j] -
                                              CNT[midx]->dDepart_c_r[i][j];
                    sCidN[nNcid]->sN[k].nRid = rid; 

#ifdef DEBUG
fprintf(stderr, " In DumpAvg() time=%f, i=%d, j=%d, cid=%d, rid=%d, N=%f, N2=%f \n", 
CTR.dCurrentTime,
i,j, cid, rid, sCidN[nNcid]->sN[k].dN,
CNT[midx]->dAccepted_c_r[i][j] - CNT[midx]->dDepart_c_r[i][j]);
fflush(stderr);
#endif
                }

                /* ----------------------------------------------- */
                /* fill in the parameter list                      */
                /* ----------------------------------------------- */
                dN = (double *)malloc( CS[cs].nNumResources[i] *
                                            sizeof(double)            );
                dBp = (double *)malloc( CS[cs].nNumResources[i] *
                                            sizeof(double)            );
                assert(dN && dBp);

                dTotalAdmission = 1.0;
                for (j=0;j < CS[cs].nNumResources[i];j++) {
                    rid = idx2rid_at(CTR.dCurrentTime, 
                                     CS[cs].sResources[i][j].nIndex);
                    for (k=0;k < CS[cs].nNumResources[i];k++) {
                        if ( rid == sCidN[nNcid]->sN[k].nRid ) {
                            break;
                        }
                    }
                    assert(k < CS[cs].nNumResources[i]);
                    dN[j] = sCidN[nNcid]->sN[k].dN;

                    if ( CNT[midx]->dPassed_c_r[i][j] +
                         CNT[midx]->dBlocked_c_r[i][j] == 0.0 ) {
                        dBp[j] = 0.0;
                    } else {
                        dBp[j] = CNT[midx]->dBlocked_c_r[i][j] /
                             (CNT[midx]->dPassed_c_r[i][j] +
                              CNT[midx]->dBlocked_c_r[i][j]   );
                    }
                    dTotalAdmission *= 1 - dBp[j];
                }
    
                DumpFunction(CTR.dCurrentTime, cid, &dN, &dBp,
                             dTotalAdmission, DumpParam);
#ifdef DEBUG
fprintf(stderr, "DumpFunction cid = %d, time = %f, dN[0]=%f\n\n", cid, CTR.dCurrentTime, dN[0]);
#endif
            
                /* free momories */
                free(dN);
                free(dBp);
            }
        }
    }
    /* free sCidN */
	for (i=0;i < nsCidNSize;i++) {
		free(sCidN[i]->sN);
		free(sCidN[i]);
	} 
	free(sCidN);
}
