/*
 * Performance evealuation of hierarchical routing by z-iteration.
 * 
 * Original version belongs to Catalin Popescu. 
 *
 * Modified by Sungjoon Ahn.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <memory.h>
#include <math.h>
#include <limits.h>
#include <string.h>
#include "mcmr.h"
#include "hrouting.h"
/* #include "structures.h" */


/************************************************************/
/* global variables                                         */
/************************************************************/
/* feasibility counter */
int f_s, f_r;

/* adjacency matrix */
static int nLinkMatrix[ MAX_NODES * MAX_NODES ];

/* inter classes */
unsigned int nIntClasses;
INTER_CLASS **intclasses;

/* inner classes */
unsigned int nClasses;
TRAFFIC_CLASS **classes;

/* outer classes */
unsigned int nOuter;
OUTER_CLASS *pOuter;

unsigned int nLinks;
unsigned int *nBandwidth;
double *dUtilization;

unsigned int nNodes;
int nMethod;

unsigned nLvOneNodes=0;
unsigned nLvOneLinks=0;

/* level-0 nodes' level-1 membership */ 
unsigned int *pNodeGrp;

/* level-1 nodes - contains every level-0 nodes inside */
LEVEL_ONE_NODE *pLvOneNdGrp;

/* level-0 links' endpoints level-1 membership */
LEVEL_ZERO_LINK *pLinkEptGrp; 

/* level-1 links - contains every level-0 links inside */
LEVEL_ONE_LINK *pLvOneLkGrp;

/* view array */
VIEW *view;


/************************************************************/
/* function prototypes                                      */
/************************************************************/
unsigned int getLvZeroRTSize();
unsigned int getLvOneRTSize();
unsigned int getLvZeroComCost();
unsigned int getLvOneComCost();
double avgPathLen();
double numCon();
double avgUtil(double *, unsigned int);
void GetAcrLink(unsigned int, unsigned int *, unsigned int *);
void GetAcrNode(unsigned int, unsigned int *, unsigned int *);
unsigned int GetNdIdx(unsigned int, unsigned int, VIEW *); 
unsigned int LkOneTypeZeroLookUp(unsigned int, unsigned int); 
unsigned int LkOneTypeOneLookUp(unsigned int, unsigned int, 
                                unsigned int); 
void PrintInterClass(int);
void PrintSingleInterClass(unsigned int);
void PrintView();
void PrintLevelZeroGrpId();
void PrintLevelZeroLinkEndPtr();
int PrintClasses(TRAFFIC_CLASS *);
int PrintServices(OUTER_CLASS *, int);
int PrintRoutes(OUTER_CLASS *, int);
void ComputeLvOneRoutes(); 
void ComputeLvOneRoute(unsigned int, unsigned int, unsigned int,
         unsigned int **, struct pNode_tag **, unsigned int);
void BuildClasses(); 
unsigned int Get_tplen(struct pNode_tag *);
unsigned int GetLvZeroLink(unsigned int, unsigned int);
unsigned int GetLvOneLink(unsigned int, unsigned int,
                          unsigned int, unsigned int);
void GetLvZeroNodes(unsigned int, unsigned int,
                    unsigned int, unsigned int *, unsigned int **);
void ShortestWithinGroup(unsigned int, unsigned int, unsigned int,
                         unsigned int *, unsigned int ***);
void Shortest(unsigned int *, unsigned int, 
              unsigned int, unsigned int,
              unsigned int *, unsigned int ***, 
              unsigned int *, unsigned int *,
              unsigned int);
int ReplaceLvOneNd(struct pNode_tag *, unsigned int, 
                    unsigned int *);
void ExpandLvOnePath(struct pNode_tag ***, unsigned int *, 
                     unsigned int *, unsigned int, 
                     unsigned int, unsigned int);
int dumper( double crtTime, int classID, double* dN[], double* dBlocking[], 
            double  dAccess, void * pDumpParam) ;
void m_module( double dRecompute, double dInterval );

/************************************************************/
/* function                                                 */
/************************************************************/
/* get the size of level-0 RT */
unsigned int getLvZeroRTSize()
{
	return nNodes * nNodes * nNodes;
}

/* get the size of level-1 RT */
unsigned int getLvOneRTSize()
{
	unsigned int i, sum = 0;

	for (i=0;i < nLvOneNodes;i++) {
		sum += view[i].nNodes * view[i].nNodes * view[i].nNodes;
	}

	return sum;
}

/* get the level-0 routing update communication cost */ 
unsigned int getLvZeroComCost()
{
	return 2 * nLinks * nLinks;
}

/* get the level-1 routing update communication cost */ 
unsigned int getLvOneComCost()
{
	unsigned int i, j, nrel, sum;

	/* level-0 communication */
	for (j=0, sum = 0;j < nLvOneNodes;j++) {
		/* count the number of edges contained in grp j */
		for (i=0, nrel=0;i < nLinks;i++) {
			if (pLinkEptGrp[i].s == pLinkEptGrp[i].e &&
				pLinkEptGrp[i].s == j) {
				nrel++;
			}
		}
		sum += 2 * nrel * nrel;
	}

	/* level-1 communication */
    /* count the number of edges of type 0 */ 
    for (i=0, nrel=0;i < nLvOneLinks;i++) {
    	if (pLvOneLkGrp[i].type == LINK_AGR_TYPE0) {
                nrel++;
        }
    }
    sum += 2 * nrel * nrel;

	return sum;
}

/* get average path length, each service has equal path
   equivalent to service setup time                         */
double avgPathLen()
{
	unsigned int i,j;
	double sum;

	for (i=0, sum=0.;i < nOuter;i++) {
		for (j=0;j < pOuter[i].nInnerClass;j++) {
			if (pOuter[i].pInnerClass[j]) {
				sum += pOuter[i].pInnerClass[j]->dWeight  
                     * pOuter[i].pInnerClass[j]->nHops;
			}
		}
	}

	return sum / nOuter; 
}

/* return number of connections in the network */
double numCon()
{
    unsigned int i,j;
    double sum;

    for (i=0, sum=0.;i < nOuter;i++) {
        for (j=0;j < pOuter[i].nInnerClass;j++) {
            if (pOuter[i].pInnerClass[j]) {
                sum += pOuter[i].pInnerClass[j]->dN;
            }
        }
    }

    return sum;
}

/* return average utilization of all the links */
double avgUtil(double *dU, unsigned int nLinks)
{
    unsigned int i;
    double sum;

    for (i=0, sum=0.;i < nLinks;i++) {
        sum += dU[i];
    }
    return sum / nLinks;
}

/* get aggregated acr of the level-1 link */
void GetAcrLink(unsigned int link, unsigned int *acr, unsigned int *cap)
{
	unsigned int i, j, k, sum;

	link--;

	/* get *cap */
	for (i=0,sum=0;i < pLvOneLkGrp[link].num;i++) {
		sum += nBandwidth[pLvOneLkGrp[link].nz[i]-1];	
	}   
	*cap = (double)sum / pLvOneLkGrp[link].num;

	/* get *acr */
	for (i=0,sum=0;i < pLvOneLkGrp[link].num;i++) { /* for each physical link */
		unsigned int zlink = pLvOneLkGrp[link].nz[i]-1;

    	for (j=0;j < nOuter;j++) {
    		for (k=0;k < pOuter[j].nInnerClass;k++) {
           	  if (pOuter[j].pInnerClass[k]) {
				if (pOuter[j].pInnerClass[k]->nBelong[zlink]) {
           	      sum += pOuter[j].pInnerClass[k]->dN *
           	             pOuter[j].pInnerClass[k]->nReserv;
				}
           	  }
        	}
    	}
	}
	*acr = (double)sum / pLvOneLkGrp[link].num;

	return;
}

/* get aggregated acr of the level-1 node */
void GetAcrNode(unsigned int node, unsigned int *acr, unsigned int *cap)
{
	unsigned int i, j, k, sum, nrel;
	unsigned int *pLinks;

	/* links contained compleletely in group 'node' */
	pLinks = (unsigned int *)malloc(nLinks * sizeof(unsigned int));
	assert(pLinks);   
	memset(pLinks, 0, nLinks * sizeof(unsigned int));

	for (i=0, nrel=0;i < nLinks;i++) {
		if (pLinkEptGrp[i].s == pLinkEptGrp[i].e) {
			pLinks[i] = 1;
			nrel++;
		}
	}

	/* get *cap */
	for (i=0,sum=0;i < nLinks;i++) {
		if (pLinks[i] == 1) {
			sum += nBandwidth[i];	
		}
	}
	*cap = (double)sum / nrel; 

	/* get *acr */
	for (i=0,sum=0;i < nLinks;i++) { /* for each physical link */
     if (pLinks[i] == 1) {
    	for (j=0;j < nOuter;j++) {
    		for (k=0;k < pOuter[j].nInnerClass;k++) {
           	  if (pOuter[j].pInnerClass[k] &&
           	      pOuter[j].pInnerClass[k]->nBelong[i]) {
           	      sum += pOuter[j].pInnerClass[k]->dN *
           	             pOuter[j].pInnerClass[k]->nReserv;
           	  }
        	}
    	}
     }
	}
	*acr = (double)sum / nrel; 

	free(pLinks);
	return;
}

/* return the matching index from view->pNd */
unsigned int GetNdIdx(unsigned int level, unsigned int nodenum, VIEW * v) 
{
    int i;

    if (level == LEVEL0) { /* level0 nodes */
        for (i=0;i < v->nLvZeroNodes;i++) {
            if (nodenum == v->pNd[i].id)
                return i;
        }
        printf("PANIC0: in GetNdIdx(), can't be here!!!\n");
        exit(1);
    } else if (level == LEVEL1) { /* level1 nodes */
        for (i=v->nLvZeroNodes;i < v->nNodes;i++) {
            if (nodenum ==  v->pNd[i].id)
                return i;
        }
        printf("PANIC1: in GetNdIdx(), can't be here!!!\n");
        exit(1);
    } else {
    }
}

/* print each level-1 paths */
void PrintInterClass(int n)
{
    int i,j;

    printf("PrintInterClass(%d)\n", n);
    for (i=0;i < nIntClasses;i++) {
        printf("%d -> Service:%3d, View:%3d, Source:%3d, Dest:%3d\n", i,
                intclasses[i]->serv_idx,
                intclasses[i]->viewidx,
                pOuter[ intclasses[i]->serv_idx ].nStartNode, 
                pOuter[ intclasses[i]->serv_idx ].nStopNode); 

        for (j=0;j < intclasses[i]->numpNodes;j++) {
            printf("(%3d,%3d)", intclasses[i]->pNodes[j].id, 
                    intclasses[i]->pNodes[j].lv);
        }
        printf("\nnclasses %3d\n", intclasses[i]->nclasses);
        if (n == 1 && !intclasses[i]->nclasses) { 
            printf("PANIC: no level-0 paths!\n");
        }
        for (j=0;j < intclasses[i]->nclasses; j++) {
            int l, k = *(intclasses[i]->class_idx + j);
            printf("Path %3d, len %3d: ", j, classes[k]->nHops);
            for (l=0;l < classes[k]->nHops+1;l++) {
                printf("%3d,", *(classes[k]->nPath + l));
            }
            printf("\n");
        } 

        printf("\n");
    }
}

/* print a particular interclass, printout level-0 path */
void PrintSingleInterClass(unsigned int i)
{
    int j;

    printf("PrintSingleInterClass(%d)\n", i);
    printf("Service:%3d, View:%3d, Source:%3d, Dest:%3d\n",
            intclasses[i]->serv_idx,
            intclasses[i]->viewidx,
            pOuter[ intclasses[i]->serv_idx ].nStartNode,
            pOuter[ intclasses[i]->serv_idx ].nStopNode);

    for (j=0;j < intclasses[i]->numpNodes;j++) {
        printf("(%3d,%3d)", intclasses[i]->pNodes[j].id,
                intclasses[i]->pNodes[j].lv);
    }
    printf("\nnclasses %3d\n", intclasses[i]->nclasses);
    for (j=0;j < intclasses[i]->nclasses; j++) {
        int l, k = *(intclasses[i]->class_idx + j);
        printf("Path %3d, len %3d: ", j, classes[k]->nHops);
        for (l=0;l < classes[k]->nHops+1;l++) {
            printf("%3d,", *(classes[k]->nPath + l));
        }
        printf("\n");
    } 
    printf("\n");
}

/* print view for each source */
void PrintView()
{
    int i, j, k;

    printf("PrintView()\n");
    for (i = 0;i < nLvOneNodes;i++) { /* for each level-1 node */
        printf("View :%3d\n", i);   
        printf("level-0 nodes %3d\n", view[i].nLvZeroNodes);
        for (j=0;j < view[i].nLvZeroNodes;j++) {
            printf("%3d ", view[i].pNd[j].id);
        }
        printf("\n");
        printf("level-1 nodes %3d\n", view[i].nNodes - view[i].nLvZeroNodes);
        for (;j <  view[i].nNodes;j++) {
            printf("%3d ", view[i].pNd[j].id);
        }
        printf("\n");

        /* adjacency matrix */
        /* header */    
        printf("\n        ");
        for (j=0; j < view[i].nNodes;j++) {
            printf("(%2d,%2d) ", view[i].pNd[j].id, view[i].pNd[j].lv);
        } 
        printf("\n");
        /* contents */
        for (j=0; j < view[i].nNodes;j++) {
            printf("(%2d,%2d) ", view[i].pNd[j].id, view[i].pNd[j].lv);
            for (k=0;k < view[i].nNodes;k++) {
                 printf("(%2d,%2d) ",view[i].pAj[j + k * view[i].nNodes ].id,
                                     view[i].pAj[j + k * view[i].nNodes ].lv); 
            }
            printf("\n");
        }
        printf("\n");
    }

}

/* return the index of pLvOneLkGrp , (x,y) type 
   if the pair doesn't exist, grow the memory and return the last element 
   make : x < y 
*/
unsigned int LkOneTypeZeroLookUp(unsigned int x, unsigned int y)
{
    int i, flag;
    unsigned int small, large;

    small = min(x,y);
    large = max(x,y);
    x= small;
    y = large;

    if (pLvOneLkGrp) {
        for (i=0, flag = 0; i < nLvOneLinks;i++) {
            if (pLvOneLkGrp[i].type == LINK_AGR_TYPE0) {
                if (x == pLvOneLkGrp[i].id.s0 && y == pLvOneLkGrp[i].id.e) {
                    flag = 1;
                    break;
                }
            }
        }
    }
    if (flag == 1) { /* found */
        return i;
    } else { /* allocate new space */
        LEVEL_ONE_LINK *tmp;

        tmp = (LEVEL_ONE_LINK *)malloc((nLvOneLinks+1)*sizeof(LEVEL_ONE_LINK));
        if (tmp == NULL) {
            fprintf(stderr, "Cannot allocate memory! - abort\n");
            exit(1);
        }
        if (nLvOneLinks) { /* copy previous contents */
            memcpy(tmp, pLvOneLkGrp, nLvOneLinks*sizeof(LEVEL_ONE_LINK));
            free(pLvOneLkGrp);
        }
        pLvOneLkGrp = tmp;
        tmp += nLvOneLinks;
        memset(tmp, 0, sizeof(LEVEL_ONE_LINK));

        nLvOneLinks++;
        return nLvOneLinks-1;
    } 
}
 
/* return the index of pLvOneLkGrp , (x.i,y) type 
   if the pair doesn't exist, return the last and increment num 
*/
unsigned int LkOneTypeOneLookUp(unsigned int x, unsigned int x0, unsigned int y)
{
    int i, flag;

    if (pLvOneLkGrp) {
        for (i=0, flag = 0; i < nLvOneLinks;i++) {
            if (pLvOneLkGrp[i].type == LINK_AGR_TYPE1) {
                if (x == pLvOneLkGrp[i].id.s0 && 
                    x0 == pLvOneLkGrp[i].id.s1 && 
                    y == pLvOneLkGrp[i].id.e)    {
                      flag = 1;
                      break;
                }
            }
        }
    }
    if (flag == 1) { /* found */
        return i;
    } else { /* allocate new space */
        LEVEL_ONE_LINK *tmp;

        tmp = (LEVEL_ONE_LINK *)malloc((nLvOneLinks+1)*sizeof(LEVEL_ONE_LINK));
        if (tmp == NULL) {
            fprintf(stderr, "Cannot allocate memory! - abort\n");
            exit(1);
        }
        if (nLvOneLinks) { /* copy previous contents */
            memcpy(tmp, pLvOneLkGrp, nLvOneLinks*sizeof(LEVEL_ONE_LINK));
            free(pLvOneLkGrp);
        }
        pLvOneLkGrp = tmp;
        tmp += nLvOneLinks;
        memset(tmp, 0, sizeof(LEVEL_ONE_LINK));

        nLvOneLinks++;
        return nLvOneLinks-1;
    } 
}
 
/* print out global pNodeGrp & pLvOneNdGrp */
void PrintLevelZeroGrpId()
{
    int i, j;

    printf("PrintLevelZeroGrpId()\n");
    /* pNodeGrp */
    printf("pNodeGrp[]\n");
    for (i=0;i < nNodes;i++) {
        printf("node %2d is in group %2d\n", i, pNodeGrp[i]); 
    }
    printf("\n");

    /* pLvOneNdGrp */
    printf("pLvOneNdGrp[]\n");
    for (i=0;i < nLvOneNodes;i++) {
        printf("grp %2d has %3d level-0 nodes:", i, pLvOneNdGrp[i].num); 
        for (j=0;j < pLvOneNdGrp[i].num;j++) {
            printf("%2d ", pLvOneNdGrp[i].nz[j]);
        }
        printf("\n");
    }
    printf("\n");
}

/* print out global pLinkEptGrp & pLvOneLkGrp */
void PrintLevelZeroLinkEndPtr()
{
    int i, j;

    printf("PrintLevelZeroLinkEndPtr()\n");
    for (i=0;i < nLinks;i++) {
        /* note that link ids start from 1 */
        printf("link %2d  = (%3d.%3d, %3d.%3d)\n", 
                i+1, pLinkEptGrp[i].s, pLinkEptGrp[i].s0,
                     pLinkEptGrp[i].e, pLinkEptGrp[i].e0); 
    }
    printf("\n");

    /* pLvOneLkGrp */
    printf("pLvOneLkGrp[]\n");
    for (i=0;i < nLvOneLinks;i++) {
        printf("level-1 link %2d has type %d and %3d level-0 links:\n", 
                i+1, pLvOneLkGrp[i].type, pLvOneLkGrp[i].num);
        if (pLvOneLkGrp[i].type == LINK_AGR_TYPE0) {
            printf("Name : (%3d,%3d) ", 
                    pLvOneLkGrp[i].id.s0, pLvOneLkGrp[i].id.e);
        } else {
            printf("Name : (%3d.%3d,%3d) ", pLvOneLkGrp[i].id.s0, 
                   pLvOneLkGrp[i].id.s1, pLvOneLkGrp[i].id.e);
        }
        printf(", Physical Links:");
        for (j=0;j < pLvOneLkGrp[i].num;j++) {
            printf("%3d ", pLvOneLkGrp[i].nz[j]);
        }
        printf("\n");
    }
    printf("\n");
}

int PrintClasses(TRAFFIC_CLASS *cls)
{
        int i;

        if (cls == NULL) { 
                printf("This class has been cut!!\n");
                return 0;
        }
        printf("\n");
        printf("lambda : %lf, vlambda : %lf, mu : %lf, Wp : %lf, OWp : %lf\n",
         cls->dL, cls->dLTime, cls->dM, cls->dWeight, cls->dOldWeight);
        printf("nReserv : %3d, nHops : %3d, nLinks : %3d\n",
         cls->nReserv, cls->nHops, cls->nLinks);


        printf("nBelong[]: ");
        for (i=0;i < cls->nLinks;i++) {
                printf("%2d,", cls->nBelong[i]);
        }
        printf("\n");
        
/*
        printf("dBlocking[]: ");
        for (i=0;i < cls->nLinks;i++) {
                printf("%lf,", cls->dBlocking[i]);
        }
        printf("\n");
        
        printf("dZ[]: ");
        for (i=0;i < cls->nLinks;i++) {
                printf("%lf,", cls->dZ[i]);
        }
        printf("\n");
        
        printf("dZOld[]: ");
        for (i=0;i < cls->nLinks;i++) {
                printf("%lf,", cls->dZOld[i]);
        }
        printf("\n");
*/

        printf("dN : %lf, dNOld : %lf, dTotalBlocking : %lf \n",
         cls->dN, cls->dNOld, cls->dTotalBlocking);

        printf("\n");
}

int PrintServices(OUTER_CLASS *serv, int nServices)
{
        int i, j;

        printf("PrintServices()\n");
        for (i=0;i < nServices;i++) {
            printf("================= Service %2d =================\n", i);
            printf("lambda : %lf, vlambda : %lf, mu : %lf, Wp : %lf\n", 
               serv[i].dL, serv[i].dLTime, serv[i].dM, serv[i].dTotalWeight);
            printf("Start : %2d, Stop : %2d, nReserv : %3d\n", 
               serv[i].nStartNode, serv[i].nStopNode, serv[i].nReserv);
            printf("nMinHops %3d\n", serv[i].nMinHops);
            for (j=0;j < serv[i].nInnerClass;j++) {
             printf("Class : %2d, pInnerClass:%d\n", j, serv[i].pInnerClass[j]);
             PrintClasses(serv[i].pInnerClass[j]);
            }
            printf("nInterClass : %2d\n", serv[i].nInterClass);
            printf("===============================================\n\n");
        }
}

int PrintRoutes(OUTER_CLASS *serv, int nServices)
{
        int i, j, k;

        for (i=0;i < nServices;i++) {
            printf("================= Service %2d =================\n", i);
            printf("Start : %2d, Stop : %2d, nReserv : %3d\n",
               serv[i].nStartNode, serv[i].nStopNode, serv[i].nReserv);
            printf("nMinHops %3d\n", serv[i].nMinHops);
            for (j=0;j < serv[i].nInnerClass;j++) {
                if (serv[i].pInnerClass[j]) {
                    printf("nBelong[]: ");
                    for (k=0;k < serv[i].pInnerClass[j]->nLinks;k++) {
                        /*printf("%2d,", serv[i].pInnerClass[j]->nBelong[k]);*/
                        if (serv[i].pInnerClass[j]->nBelong[k])
                            printf("%2d,", k+1);
                    }
                    printf("\n");
                }
            }
        }
}

/* returns length of a path */
unsigned int Get_tplen(struct pNode_tag *p)
{
    int i=0;

    while (p[i++].id != UINT_MAX); 
    return i-1;
}

/* return the level-0 link id */
unsigned int GetLvZeroLink(unsigned int n1, unsigned int n2)
{
	int i;

	for (i=0;i < nLinks;i++) {
		if ((pLinkEptGrp[i].s0 == n1 && pLinkEptGrp[i].e0 == n2) ||
		    (pLinkEptGrp[i].s0 == n2 && pLinkEptGrp[i].e0 == n1))   {
			return i+1;
		} 
	}
	assert(0);
}

/* return the level-1 link id */
unsigned int GetLvOneLink(unsigned int id1, unsigned int lv1, 
                          unsigned int id2, unsigned int lv2)
{
    int i;

    if (lv1 == lv2 && lv1 == LEVEL0) {
        printf("In GetLvOneLink(): one of the node should be level-1\n");
        exit(1);
    }

    for (i=0;i < nLvOneLinks;i++) {
        if (lv1 == LEVEL1 && lv2 == LEVEL1 && 
            pLvOneLkGrp[i].type == LINK_AGR_TYPE0) {
            if ((id1==pLvOneLkGrp[i].id.s0 && id2==pLvOneLkGrp[i].id.e) || 
                (id2==pLvOneLkGrp[i].id.s0 && id1==pLvOneLkGrp[i].id.e))
            return i+1;
        }
        if (lv1 == LEVEL1 && lv2 == LEVEL0 && 
            pLvOneLkGrp[i].type == LINK_AGR_TYPE1) {
            if (id1==pLvOneLkGrp[i].id.e && id2==pLvOneLkGrp[i].id.s1)  
            return i+1;
        }
        if (lv1 == LEVEL0 && lv2 == LEVEL1 && 
            pLvOneLkGrp[i].type == LINK_AGR_TYPE1) {
            if (id2==pLvOneLkGrp[i].id.e && id1==pLvOneLkGrp[i].id.s1)  
            return i+1;
        }
    }
}

/* return level-0 nodes which are on the opposite side of  
   the given combination */
void GetLvZeroNodes(unsigned int LvOneLink, unsigned int id,
                    unsigned int lv, unsigned int *num, unsigned int **ptr)
{
    unsigned int nlink, grp, *pVisited;
    int i,j;

    LvOneLink--;

    if (lv == LEVEL0) {
        grp = pLvOneLkGrp[LvOneLink].id.e;   
    } else { /* lv = LEVEL1 */
        if (id == pLvOneLkGrp[LvOneLink].id.e) { /* take the opposite one */
            grp = pLvOneLkGrp[LvOneLink].id.s0;
        }
        else { 
            grp = pLvOneLkGrp[LvOneLink].id.e; 
        }
    }

    nlink = pLvOneLkGrp[LvOneLink].num;  
    *ptr = (unsigned int *)malloc(sizeof(unsigned int) * nlink);
    if (*ptr == NULL) {
        printf("Cannot allocate memory - GetLvZeroNodes()\n");
        exit(1);
    }

    /* eliminate redundant nodes */
    pVisited = (unsigned int *)malloc(sizeof(unsigned int) * nNodes);
    if (pVisited == NULL) {
        printf("Cannot allocate memory - GetLvZeroNodes()\n");
        exit(1);
    }
    memset(pVisited, 0, sizeof(unsigned int) * nNodes);

    *num = 0;
    for (i=0,j=0;i < nlink;i++) {
        unsigned int pl;

        pl = pLvOneLkGrp[LvOneLink].nz[i];
        pl--;

        if (grp == pLinkEptGrp[pl].s) {
            if (!pVisited[pLinkEptGrp[pl].s0]) {
                *(*ptr + j) = pLinkEptGrp[pl].s0;
                (*num)++;
                pVisited[pLinkEptGrp[pl].s0] = 1;
                j++;
            }
        } else if (grp == pLinkEptGrp[pl].e) {
            if (!pVisited[pLinkEptGrp[pl].e0]) {
                *(*ptr + j) = pLinkEptGrp[pl].e0;
                (*num)++;
                pVisited[pLinkEptGrp[pl].e0] = 1;
                j++;
            }
        } else {
            printf("PANIC:This can't happen! in GetLvZeroNodes()\n");
            exit(1);
        }
    }
    free(pVisited);
}

/* get the shortest path within a group, return numsp and ptrsp 
   Note that the end of the path is marked by UINT_MAX          
   also that it's possible to have n1==n2 */
void ShortestWithinGroup(unsigned int gid, unsigned int n1, unsigned int n2,
                         unsigned int *numsp, unsigned int ***ptrsp) 
{
    unsigned int minhop = UINT_MAX;
    unsigned int i, j ,cnt = 0, srclen = 0;
    unsigned int **tmp;
    unsigned int *pVisited, *pNodes;

    /* take care of the case where n1 == n2 */
    if (n1 == n2) {
        *numsp = 1;
        tmp = (unsigned int **)malloc(sizeof(unsigned int *));
        pNodes = (unsigned int *)malloc(nNodes * sizeof(unsigned int));
        if (tmp == NULL || pNodes == NULL) {
            printf("Can't allocate memory-in ShortestWithinGroup()\n");
            exit(1);
        }
        pNodes[0] = n1;
        pNodes[1] = UINT_MAX;
        tmp[0] = pNodes;
        *ptrsp = tmp;

        return;
    }

    /* initialize */
    *numsp = 0;

    pVisited = (unsigned int *)malloc(nNodes*sizeof(unsigned int));
    if (pVisited == NULL) {
        printf("Can't allocate memory-in ShortestWithinGroup()\n");
        exit(1);
    }
    memset(pVisited, 0, nNodes*sizeof(unsigned int));
    pVisited[GetNdIdx(LEVEL0, n1, &view[gid])] = 1;
    pNodes = (unsigned int *)malloc(nNodes*sizeof(unsigned int));
    if (pVisited == NULL) {
        printf("Can't allocate memory-in ShortestWithinGroup()\n");
        exit(1);
    }
    pNodes[0] = n1;
    Shortest(&minhop, gid, n1, n2, numsp, ptrsp, pVisited, pNodes, 0);

    /* determine minimum hop length */
    for (i=0;i < *numsp;i++) {
        while ( *(*(*ptrsp + i) + srclen) != UINT_MAX) srclen++;
        if (minhop > srclen-1) {
            minhop = srclen-1;
        } 
    }

    /* count paths whose lengths <= minhop + 1 */
    for (i=0;i < *numsp;i++) {
        while ( *(*(*ptrsp + i) + srclen) != UINT_MAX) srclen++;
        if (minhop + 1 >= srclen-1) {
            cnt++;
        } 
    }

    /* save those */ 
    tmp = (unsigned int **)malloc(cnt*sizeof(unsigned int *));
    if (tmp == NULL) {
        printf("Can't allocate memory-in ShortestWithinGroup()\n");
        exit(1);
    }
    
    for (i=0,j=0;i < *numsp;i++) {
        while ( *(*(*ptrsp + i) + srclen) != UINT_MAX) srclen++;
        if (minhop + 1 >= srclen-1) {
            tmp[j++] = *(*ptrsp + i);
        } else { /* discard */
            free(*(*ptrsp + i));
        } 
    }
    free(*ptrsp);
    *ptrsp = tmp;
    *numsp = cnt;
    
    free(pVisited);
    free(pNodes);
}

/* actual work of ShortestWithinGroup() is done here */
void Shortest(unsigned int *minhop, unsigned int gid, 
              unsigned int n1, unsigned int n2,
              unsigned int *numsp, unsigned int ***ptrsp, 
              unsigned int *pVisited, unsigned int *pNodes,
              unsigned int fringe)
{
    if (*minhop != UINT_MAX && fringe > *minhop+1) {
        return;
    }

    /* check if the path has been completed */
    if (n2 == pNodes[fringe]) { /* copy the path to the list structure */
        unsigned int **sparr;
        unsigned int *sp;

        sparr = (unsigned int **)malloc((*numsp+1) * sizeof(unsigned int *));
        sp = (unsigned int *)malloc(nNodes * sizeof(unsigned int));
        if (sparr == NULL || sp == NULL) {
            printf("Can't allocate memory-in Shortest()\n");
            exit(1);
        }
        memset(sp, 0xff, nNodes * sizeof(unsigned int));    
        memcpy(sp, pNodes, (fringe+1) * sizeof(unsigned int));

        if (*numsp) {
            memcpy(sparr, *ptrsp, *numsp * sizeof(unsigned int *));
            free(*ptrsp);
        }
        sparr[(*numsp)++] = sp;
        *ptrsp = sparr; 

        if (fringe < *minhop) { /* fringe = path length */
            *minhop = fringe;
        }
    } else { /* find new nodes */
        unsigned int i;

        for (i=0;i < view[gid].nLvZeroNodes;i++) {
            unsigned int tmp1;

            tmp1 = GetNdIdx(LEVEL0, pNodes[fringe], &view[gid]);
            if (view[gid].pAj[ tmp1 + i * view[gid].nNodes ].id &&
                !pVisited[i]) { /* if connected but not yet visited */

                pVisited[i] = 1;
                pNodes[fringe + 1] = view[gid].pNd[i].id;

                Shortest(minhop, gid, n1, n2, numsp, ptrsp, pVisited, pNodes, fringe+1);

                /* roll-back */
                pVisited[i] = 0;
            }
        }
    }
}

/* replace logical node at dest[i] with subpath in src,
   return number of replacements                      */
int ReplaceLvOneNd(struct pNode_tag *dest, unsigned int pos, 
                    unsigned int *src)
{
    int i, srclen=0;

    /* get the length of the source */
    while (src[srclen] != UINT_MAX) srclen++;

    /* shift the portion that is after i  by srclen-1 */
    for (i=Get_tplen(dest);i > pos;i--) {
        dest[i + srclen - 1].id = dest[i].id;
        dest[i + srclen - 1].lv = dest[i].lv;
    }
    for (i=pos;i < pos + srclen;i++) {
        dest[i].id = src[i-pos];  
        dest[i].lv = LEVEL0;
    }
    return srclen;
}

/* expand the level-1 path into multiple level-0 paths,
   expand a level-1 node at a time                       */
void ExpandLvOnePath(struct pNode_tag ***tparr, unsigned int *numtp, 
                     unsigned int *minhop, unsigned int stopnode, 
                     unsigned int pos, unsigned int cur)
{
    struct pNode_tag *t, *tback;
    int i, j, k, l, delta;
    unsigned int lo1, lo2, numlv1, numlv2;
    unsigned int *lv1Nds1, *lv1Nds2;
    unsigned int numsp;
    unsigned int **ptrsp;

    t = *(*tparr + cur);
    i = pos;
    while (t[i].id != UINT_MAX && t[i].lv == 0) i++;
    if (t[i].id == UINT_MAX && t[i-1].id == stopnode &&
        t[i-1].lv == LEVEL0) { /* expanded everything already */

        /* i = hop length */
        if (*minhop > i) {
            *minhop = i;
        }
        return;
    }

    /* make backup */
    tback = (struct pNode_tag *)malloc(nNodes * sizeof(struct pNode_tag));
    if (!tback) {
        printf("Cannot allocate memory!\n");
        exit(1);
    }
    memcpy(tback, t, nNodes * sizeof(struct pNode_tag));

    /* i points to the first level-1 node */
    /* note, t[i-1]:level0, t[i]:level1 */
    lo1 = GetLvOneLink(t[i-1].id, t[i-1].lv, t[i].id, t[i].lv); 
    GetLvZeroNodes(lo1, t[i-1].id, t[i-1].lv, &numlv1, &lv1Nds1);

    if (t[i+1].id != UINT_MAX) {
        /* note, t[i]:level1, t[i+1]:level? */
        lo2 = GetLvOneLink(t[i].id, t[i].lv, t[i+1].id, t[i+1].lv); 
        GetLvZeroNodes(lo2, t[i+1].id, t[i+1].lv, &numlv2, &lv1Nds2);
    } else { /* t[i+1].id is the last node */ 
        numlv2 = 1;
        lv1Nds2 = (unsigned int *)malloc(sizeof(unsigned int));
        if (lv1Nds2 == NULL) {
            printf("Can't allocate memory- in ExpandLvOnePath()\n");
            exit(1);
        }
        *lv1Nds2 = stopnode;
    }

    /* for each level-0 end point, obtain multiple shortest path or (+1) */
    for (j=0;j < numlv1;j++) {
        for (k=0;k < numlv2;k++) {
            ShortestWithinGroup(tback[i].id, lv1Nds1[j], lv1Nds2[k], &numsp, &ptrsp);   
            for (l=0;l < numsp;l++) {
              if (j || k || l) {/* have to allocate a new space,from 2nd path */
                struct pNode_tag *tmp;
                struct pNode_tag **tmparr;

                tmp = (struct pNode_tag *)malloc(nNodes * sizeof(struct pNode_tag));
                if (tmp == NULL) {
                    printf("Cannot allocate memory!\n");
                    exit(1);
                }
                memcpy(tmp, tback, nNodes * sizeof(struct pNode_tag));

                tmparr = (struct pNode_tag **)malloc((*numtp+1) * sizeof(struct pNode_tag *));
                if (tmparr == NULL) {
                    printf("Cannot allocate memory!\n");
                    exit(1);
                }
                if (*numtp) {
                    memcpy(tmparr, *tparr, *numtp * sizeof(struct pNode_tag *));
                    free(*tparr);
                }
                tmparr[*numtp] = tmp;
                (*numtp)++;
                *tparr = tmparr;
                
                delta = ReplaceLvOneNd(tmp, i, ptrsp[l]);
                
                /* recurse */
                ExpandLvOnePath(tparr, numtp, minhop, stopnode, i+delta, *numtp-1);
              } else { /* don't need to allocate a space */

                delta = ReplaceLvOneNd(t, i, ptrsp[l]);
                ExpandLvOnePath(tparr, numtp, minhop, stopnode, i+delta, cur);

              }
            }
            free(ptrsp);
        }   
    } 
    free(lv1Nds1);
    free(lv1Nds2);
    free(tback);
}
 
/* This function constructs all the level-0 classes */
void  BuildClasses() 
{
    int i,j,k;
    struct pNode_tag **tparr;
    struct pNode_tag *tp;
    unsigned int numtp, minhop;

    for (i=0;i < nIntClasses;i++) { /* for each level-1 path */

        /* allocate space for a temporary path */
        tp = (struct pNode_tag *)malloc(nNodes * sizeof(struct pNode_tag));
        tparr = (struct pNode_tag **)malloc(sizeof(struct pNode_tag*));
        if (tp == NULL || tparr == NULL) {
            printf("Cannot allocate memory\n");
            exit(1);
        } 
            /* if id == UINT_MAX it's end of the path */ 
        memset(tp, 0xff, nNodes * sizeof(struct pNode_tag)); 
        *tparr = tp; 
        numtp = 1;
        minhop = UINT_MAX;

        /* copy level-1 path to tp */
        memcpy(tp, intclasses[i]->pNodes, intclasses[i]->numpNodes *
               sizeof(struct pNode_tag));  

        /* call the function to expand this path */
        ExpandLvOnePath(&tparr, &numtp, &minhop, 
                        pOuter[ intclasses[i]->serv_idx ].nStopNode, 0, 0);

	assert(numtp);

        /* store the result to InnerClasses */
        for (j=0;j < numtp;j++) {
            TRAFFIC_CLASS *ic;
            TRAFFIC_CLASS **icp;
            unsigned int *cpp;
            unsigned int sv = intclasses[i]->serv_idx;

            /*if (Get_tplen(tparr[j]) <= minhop + 1) { /* I'm wondering ... */
	    if (1) {
                /* allocate a class, fill it up */
                ic = (TRAFFIC_CLASS *)malloc(sizeof(TRAFFIC_CLASS));
                if (ic == NULL) {
                    printf("Cannot allocate memory\n");
                    exit(1);
                }
                ic->dL = pOuter[intclasses[i]->serv_idx ].dL;
                ic->dM = pOuter[intclasses[i]->serv_idx ].dM;
                ic->dLTime = pOuter[intclasses[i]->serv_idx ].dLTime;
                ic->nReserv = pOuter[intclasses[i]->serv_idx ].nReserv;
                ic->dWeight = ic->dOldWeight = 0; 

                ic->nBelong = (unsigned int *)malloc(sizeof(unsigned int) *
                                                     nLinks);
                ic->dBlocking = (double *)malloc(sizeof(double) * nLinks);
                ic->dZ = (double *)malloc(sizeof(double) * nLinks);
                ic->dZOld = (double *)malloc(sizeof(double) * nLinks);
                ic->nLinks = nLinks; 
                if (!ic->nBelong || !ic->dBlocking || !ic->dZ || !ic->dZOld) {
                    printf("Cannot allocate memory\n");
                    exit(1);
                }
                memset(ic->nBelong, 0, sizeof(unsigned int) * nLinks);
                memset(ic->dBlocking, 0, sizeof(double) * nLinks);
        
                ic->dN = 0;
                ic->dNOld = 0;
                ic->nHops = Get_tplen(tparr[j])-1;

                /* fill in nBelong - check the adjacency matrix!!! */
                for (k = 1;k < Get_tplen(tparr[j]);k++) {
                    struct pNode_tag *s;

                    s = tparr[j];
                    ic->nBelong[ nLinkMatrix[ s[k-1].id + s[k].id * MAX_NODES ] -1 ] = 1;
                } 

                ic->nPath = (unsigned int *)malloc(sizeof(unsigned int) *
                                                     nLinks);
                if (!ic->nPath) {
                    printf("Cannot allocate memory\n");
                    exit(1);
                }
                /* fill in Path */
                for (k=0;k < ic->nHops+1;k++) {
                    struct pNode_tag *s;
                    s = tparr[j];

                    *(ic->nPath + k) = s[k].id; 
                }

                /* fill in back links - change structure */
                ic->serv_idx = intclasses[i]->serv_idx;
                ic->intclass_idx = i;

                /* Link Globally */
                icp = (TRAFFIC_CLASS **)malloc((nClasses+1) * 
                                                sizeof(TRAFFIC_CLASS *));
                if (icp == NULL) {
                    printf("Cannot allocate memory! -Abort\n");
                    exit(1);
                }
                if (nClasses) { /* copy previous contents */
                    memcpy(icp, classes, nClasses * sizeof(TRAFFIC_CLASS *));
                    free(classes);
                }
                icp[nClasses++] = ic;  
                classes = icp;

                /* link to InterClass */
                cpp = (unsigned int *)malloc((intclasses[i]->nclasses+1) *
                                             sizeof(unsigned int));     
                if (cpp == NULL) {
                    printf("Cannot allocate memory! -Abort\n"); 
                    exit(1);
                }
                if (intclasses[i]->nclasses) { /* copy previous contents */
                    memcpy(cpp, intclasses[i]->class_idx, 
                           intclasses[i]->nclasses * sizeof(unsigned int));
                    free(intclasses[i]->class_idx);
                }
                cpp[intclasses[i]->nclasses++] = nClasses-1; /* check !!! */
                intclasses[i]->class_idx = cpp;

                /* link to OuterClass, i.e., service class */
                icp = (TRAFFIC_CLASS **)malloc((pOuter[sv].nInnerClass+1) * 
                                                sizeof(TRAFFIC_CLASS *));
                if (icp == NULL) {
                    printf("Cannot allocate memory! -Abort\n");
                    exit(1);
                }
                if (pOuter[sv].nInnerClass) {
                    memcpy(icp, pOuter[sv].pInnerClass, 
                           pOuter[sv].nInnerClass * sizeof(TRAFFIC_CLASS *));
                    free(pOuter[sv].pInnerClass);
                }
                icp[pOuter[sv].nInnerClass++] = ic;  
                pOuter[sv].pInnerClass = icp;

                /* pOuter[sv].nMinHops contains number of nodes en route */
                if (pOuter[sv].nMinHops > minhop-1)
                    pOuter[sv].nMinHops = minhop-1;
            }
        }
        
        /* free the allocated space */
        for (j=0;j < numtp;j++) {
            free(tparr[j]);
        }
        free(tparr);
    }
}

/* This function computes level-1 routes for each service */
void  ComputeLvOneRoutes() 
{
    int i,j, k, l;
    INTER_CLASS **icp;
    unsigned int *pVisited0;
    struct pNode_tag *pNodes0; 
    
    
    for (i=0;i < nOuter;i++) {
        ComputeLvOneRoute(i, 0, nNodes, &pVisited0, &pNodes0, 0);

        /* free temporary data structures */
        free(pVisited0);
        free(pNodes0);

        /* store it into global array
           ELIMINATE USELESS PATHS!!! */
        k = pOuter[i].nInterClass;
        for (j=0;j < k;j++) {
            if (pOuter[i].pInterClass[j]->numpNodes - 1 > pOuter[i].nLvOneMinHops + 1) {
                free(pOuter[i].pInterClass[j]);
                pOuter[i].pInterClass[j] = NULL;
                pOuter[i].nInterClass--;
            } else { /* store it */
                icp = (INTER_CLASS **)malloc((nIntClasses+1) * sizeof(INTER_CLASS *));
                if (icp == NULL) {
                    printf("Cannot allocate memory -- Abort\n");
                    exit(1);
                }
                if (nIntClasses) { /* copy previous contents */
                    for (l=0;l < nIntClasses;l++) {
                        icp[l] = intclasses[l];  
                    }
                    free(intclasses);
                }
                icp[nIntClasses++] = pOuter[i].pInterClass[j];
                intclasses = icp;
            }
        }
    } 
}

/* This function computes level-1 routes for the given service 
   should fill in the following:

   nInterClass, pInterClass for the corresponding service(pOuter)
   set of INTER_CLASSES for each service(pOuter)
   intclasses, nIntClasses (Global)
*/
void  ComputeLvOneRoute(unsigned int sv, unsigned int vn, unsigned int fringe, 
                        unsigned int **pVisited0, 
                        struct pNode_tag **pNodes0, unsigned int stopnode)
{
    int done;
    unsigned int *pVisited;
    struct pNode_tag *pNodes; 

    /* based upon the adjacency matrix of view[vn], find the shortest paths */
    /* find only minpath + i                                                */

    if (fringe > pOuter[sv].nLvOneMinHops+1) { /* drop, too long already */
        return;
    }

    if (fringe == nNodes) { /* called the first time */

        /* fetch the corresponding view */
        vn = pNodeGrp[ pOuter[sv].nStartNode ];

        /* set up the stop node correctly */
        if (pNodeGrp[ pOuter[sv].nStartNode ] == pNodeGrp[ pOuter[sv].nStopNode ]) {
            stopnode = GetNdIdx(LEVEL0, pOuter[sv].nStopNode, &view[vn]); 
        } else { /* level 1 noded */
            stopnode = GetNdIdx(LEVEL1, pNodeGrp[ pOuter[sv].nStopNode ], &view[vn]);
        }

        /* allocate necesary space */
        /* visited : to prevent cycling */
        pVisited = (unsigned int *)malloc(view[vn].nNodes * sizeof(unsigned int));
        if (pVisited == NULL) {
            printf("Cannot allocate memory -- Abort\n");
            exit(1);
        }
        memset(pVisited, 0, view[vn].nNodes * sizeof(unsigned int));

        /* pNodes : to store the path up to the current time */
        pNodes = (struct pNode_tag *)malloc(view[vn].nNodes * sizeof(struct pNode_tag));
        if (pNodes == NULL) {
            printf("Cannot allocate memory -- Abort\n");
            exit(1);
        }
        memset(pNodes, 0, view[vn].nNodes * sizeof(struct pNode_tag));
        pNodes[0].id = pOuter[sv].nStartNode; 
        pNodes[0].lv = 0; 
        pVisited[0] = 1; 
        fringe = 0;

        /* to free the structures later */
        *pNodes0 = pNodes;
        *pVisited0 = pVisited;

        /* mark the first node, it's always level 0 */
        pVisited[ GetNdIdx(LEVEL0, pOuter[sv].nStartNode, &view[vn]) ]; 

    } else {
        pVisited = *pVisited0;
        pNodes = *pNodes0;
    }

    /* check if the path has been constructed */
    if (stopnode == GetNdIdx(pNodes[fringe].lv, pNodes[fringe].id, &view[vn])) {
        done = 1;
    } else {
        done = 0;
    }

    if (!done) { /* if we didn't reach the stopnode, do DFS */
        unsigned int i;

        for (i=0;i < view[vn].nNodes;i++) {
            unsigned int tmp1;

            tmp1 = GetNdIdx(pNodes[fringe].lv, pNodes[fringe].id, &view[vn]); 
            if (view[vn].pAj[ tmp1 + i * view[vn].nNodes ].id && /* (tmp1, i) */
                !pVisited[i]) {

                pVisited[i] = 1;
                pNodes[fringe + 1].id = view[vn].pNd[i].id;
                pNodes[fringe + 1].lv = view[vn].pNd[i].lv;
                
                ComputeLvOneRoute(sv, vn, fringe+1, &pVisited, &pNodes, stopnode);

                /* roll-back */
                pVisited[i] = 0;
            }   
        }
    } else { /* store the path in the interclass structure */ 
        INTER_CLASS *ic;
        INTER_CLASS **icp;

        /* allocate an inter-class space and copy necessary stuff */
        ic = (INTER_CLASS *)malloc(sizeof(INTER_CLASS));
        if (ic == NULL) {
            printf("Cannot allocate memory -- Abort\n");
            exit(1);
        }
        ic->viewidx = vn;
        ic->numpNodes = fringe+1;
        ic->pNodes = (struct pNode_tag *)malloc((fringe+1) * sizeof(struct pNode_tag));
        if (ic->pNodes == NULL) {
            printf("Cannot allocate memory -- Abort\n");
            exit(1);
        }
        memcpy(ic->pNodes, pNodes, (fringe+1) * sizeof(struct pNode_tag));
        ic->serv_idx = sv; 
        ic->nclasses = 0;

        /* link this into the service array */ 
        icp = (INTER_CLASS **)malloc((pOuter[sv].nInterClass+1) * sizeof(INTER_CLASS *));
        if (icp == NULL) {
            printf("Cannot allocate memory -- Abort\n");
            exit(1);
        }
        if (pOuter[sv].nInterClass) { /* copy previous contents */
            int i;
            for (i=0;i < pOuter[sv].nInterClass;i++) {
                icp[i] = pOuter[sv].pInterClass[i]; 
            }
            free(pOuter[sv].pInterClass);
        }
        icp[ pOuter[sv].nInterClass ] = ic;
        pOuter[sv].nInterClass++;
        pOuter[sv].pInterClass = icp;
        if (fringe < pOuter[sv].nLvOneMinHops) { /* fringe = path length */ 
            pOuter[sv].nLvOneMinHops = fringe; 
        }

    }
}

/*
 * This funtion, is recursive one, and will compute all the routes 
 * from nNodes[0] to nStopNode. But only routes shorter than minhop + i 
 * will be stored. The routes will be stored in the pClass traffic class.
 *
 * Input :
 *   int pMatrix[] - the topology of the network
 *   OUTER_CLASS* pClass - the class we want to build the routes
 *   unsigned int nNodes - the number of nodes in the network
 *   unsigned int nCrtNode - the numbers of nodes in path till now.
 *   int pVisited[] - an array to prevent circular paths
 *   unsigned int pNodes[] - the path till now.
 *   unsigned int nLinks - the total number of links in the network.
 */
void ComposeRoute( int pMatrix[], OUTER_CLASS* pClass, 
                   unsigned int nNodes, unsigned int nCrtNode,
                   int pVisited[], unsigned int pNodes[], unsigned int nLinks
                  ) {  

  int nIndex;

  /* Already too far.. */
  if( nCrtNode  > pClass->nMinHops + 1  )
    return;

  if( pClass->nStopNode == pNodes[ nCrtNode ] ) {
    /* ---------------------------------- */
    /* the route reached the destination! */ 
    /* ---------------------------------- */

    TRAFFIC_CLASS *pNewClass; /* to hold new class */
    /* to hold existing routes + new route */
    TRAFFIC_CLASS **pNewClasses = ( TRAFFIC_CLASS ** ) malloc( sizeof( TRAFFIC_CLASS * ) * ( pClass->nInnerClass + 1 )); /* 1 bigger than current # routes */

    if( pClass->nInnerClass > 0 ) {
    /* if is not the first route of the SERVICE, copy the other routes */
    /* then free the previous memory chunk                             */
      assert( pClass->pInnerClass != NULL );
      memcpy( pNewClasses, pClass->pInnerClass, 
              pClass->nInnerClass * sizeof( TRAFFIC_CLASS * ) );
      free( pClass->pInnerClass );
    }
    
    /* Create and initialize the new Class */
    pNewClass = ( TRAFFIC_CLASS * ) malloc( sizeof( TRAFFIC_CLASS ));

    pNewClass->dL = pClass->dL;
    pNewClass->dM = pClass->dM;
    pNewClass->dLTime = pClass->dLTime;
    pNewClass->nReserv = pClass->nReserv;
    pNewClass->dWeight = pNewClass->dOldWeight = 0;
    pNewClass->nBelong = ( unsigned int *) malloc( sizeof( unsigned int ) * nLinks );
    pNewClass->dBlocking = ( double *) malloc( sizeof( double ) * nLinks );
    pNewClass->dZ = ( double *) malloc( sizeof( double ) * nLinks );
    pNewClass->dZOld = ( double *) malloc( sizeof( double ) * nLinks );
    pNewClass->nLinks = nLinks;
    memset( pNewClass->nBelong, 0, sizeof( unsigned int ) * nLinks );

    memset( pNewClass->dBlocking, 0, sizeof( double ) * nLinks );
    pNewClass->dN = 0;
    pNewClass->dNOld = 0;

    pNewClass->nHops = nCrtNode;    /* path lengths : number of hops */

    /* record the path, nBelong <- pMatrix[pNodes[nIndex-1], pNodes[nIndex]] */ 
    /* that is, record the link number or ID starting from 1 */
    for( nIndex = 1; nIndex <= nCrtNode; nIndex ++ )
      pNewClass->nBelong[ pMatrix[ pNodes[ nIndex - 1 ] +  pNodes[ nIndex ] * MAX_NODES ] - 1 ]  = 1;  
 
    pNewClasses[ pClass->nInnerClass ] = pNewClass; /* put it in the array */

    /* Update the hops */
    if( nCrtNode  < pClass->nMinHops )
      pClass->nMinHops = nCrtNode;
    
    pClass->nInnerClass ++;
    pClass->pInnerClass = pNewClasses; /* the new array */
  }
  /* expand the fringe one by one */
  else {
    for( nIndex = 0; nIndex < nNodes; nIndex ++ ) {
      /* A new possible path.. ? - make sure it's not on the path */
      if( pMatrix[ pNodes[ nCrtNode ] + nIndex * MAX_NODES ] != 0 && !pVisited[ nIndex ] ){
        /* got it ..! */
        pVisited[ nIndex ] = TRUE;
        pNodes[ nCrtNode + 1 ] = nIndex; /* add the new node to pNodes */
        
        /* next step */
        ComposeRoute( pMatrix, pClass, nNodes, nCrtNode + 1, 
                      pVisited, pNodes, nLinks );
        
        /* rollback - current node won't be included for the next ones */
        pVisited[ nIndex ] = FALSE;
      }
    }
  }
}

/*
 * This function will compute for each class all possible routes. Usses the ComposeRoute
 * function to achieve this.
 */
void ComputeClassRoutes( int pMatrix[], OUTER_CLASS pClasses[], 
                         unsigned int nNodes, unsigned int nClasses, unsigned int nLinks,
                         TRAFFIC_CLASS *** pReturnClasses, int* pNumClasses ) {
  unsigned int nIndex, nInnerIndex, nIndexClass;
  unsigned int* pNodes;
  int* pVisited;
  unsigned int nInnerClasses = 0;

  /* Allocate some buffers */
  pNodes = ( unsigned int* ) malloc( nNodes * sizeof( unsigned int ));
  if( NULL == pNodes  ) {
    fprintf( stderr, "Error: Out of memory.. \n");
    exit( -1 );
  }
  pVisited = ( int* ) malloc( nNodes * sizeof( int ));
  if( NULL == pVisited ) {
    fprintf( stderr, "Error: Out of memory.. \n");
    exit( -1 );
  }

  /* compute routes for each service */
  for( nIndex = 0; nIndex < nClasses; nIndex++ ) {
    /* nothing visited yet. */
    memset( pVisited, 0, nNodes * sizeof( int ));

    for( nInnerIndex = 0; nInnerIndex < nNodes; nInnerIndex ++ )
      pNodes[ nInnerIndex ] = MAX_NODES+1;

    pClasses[nIndex].nInnerClass = 0;
    pClasses[nIndex].pInnerClass = NULL;
    pClasses[nIndex].nMinHops = nNodes;

    /* start the route */
    pNodes[0] = pClasses[nIndex].nStartNode;
    pVisited[pClasses[nIndex].nStartNode] = TRUE;
    if( pClasses[ nIndex ].nStartNode == pClasses[ nIndex ].nStopNode )
      printf( "ERROR - class %d from %d to %d !\n", nIndex, pClasses[ nIndex ].nStartNode,
              pClasses[ nIndex ].nStopNode );

    /* get paths for each service : note that some the possible routes 
           have been cut-off by this function */
    ComposeRoute( pMatrix, &pClasses[nIndex], nNodes, 0, pVisited, pNodes,
                  nLinks );

    /* now, reduce the routes to the effective ones.. : cut off the rest */
    for( nInnerIndex = 0; nInnerIndex < pClasses[nIndex].nInnerClass; nInnerIndex ++ ) 
      if( pClasses[nIndex].pInnerClass[nInnerIndex]->nHops > pClasses[nIndex].nMinHops + 1 ) {
        /* free the inner class !! */

        free( pClasses[nIndex].pInnerClass[nInnerIndex]->dZ );
        free( pClasses[nIndex].pInnerClass[nInnerIndex]->dZOld );
        free( pClasses[nIndex].pInnerClass[nInnerIndex]->nBelong );
        free( pClasses[nIndex].pInnerClass[nInnerIndex]->dBlocking );
        
        free(  pClasses[nIndex].pInnerClass[nInnerIndex] );
        pClasses[nIndex].pInnerClass[nInnerIndex] = NULL;
      }
      else
        (*pNumClasses) ++;
  }     
  /* Now, just try to save the data .. */
  (*pReturnClasses) = ( TRAFFIC_CLASS ** ) malloc( sizeof( TRAFFIC_CLASS * ) * *pNumClasses );

  nIndexClass = 0;
  for( nIndex = 0; nIndex < nClasses; nIndex ++ )
    for( nInnerIndex = 0; nInnerIndex < pClasses[nIndex].nInnerClass; nInnerIndex ++ ) 
      if( pClasses[nIndex].pInnerClass[nInnerIndex] != NULL ) {
        (*pReturnClasses)[ nIndexClass ] = pClasses[nIndex].pInnerClass[nInnerIndex];
        nIndexClass ++;
      }  

  /* cleanup.. */
  free( pVisited );
  free( pNodes );
}

/*
 * This function will compute the utilization for all the links
 *
 * double dLinkUtilization[] - the utilization for each link
 * unsigned int nLinks - the number of links
 * TRAFFIC_CLASSES *pClasses - where to sera
 */
#ifdef GCHITTI
void ComputeUtilization( TRAFFIC_CLASS** pClasses, unsigned int nClasses,  
                         double dUtilization[], unsigned int nBandwidths[], 
                         unsigned int nLinks ) {

  unsigned int nIndex, nInnerIndex;

  /* add up effective capacities, divide it by link capacity */
  for( nIndex = 0; nIndex < nLinks; nIndex ++ ) { /* for each link */
    dUtilization[ nIndex ] = 0.0;

    for( nInnerIndex = 0; nInnerIndex < nClasses; nInnerIndex++ ) {
      if( pClasses[ nInnerIndex ]->nBelong[ nIndex ] ) { 
            /* for relevant classes, add up the usage */
            dUtilization[nIndex] += pClasses[ nInnerIndex ]->dN * 
                                    pClasses[ nInnerIndex ]->nReserv;
      }
    }
    
    /* finally .. there , divide it by link capacity */
    dUtilization[ nIndex ] = dUtilization[ nIndex ] / 
                             (double)nBandwidths[ nIndex ];
  }
}
#endif

/* Level-0 feasiblility test. 
   return 1, if the link looks feasible to source,
          0, otherwise

   The model(so does actual network) assumes that throughout an 
  update period, the source has an identical view of the network. 
  Therefore, this runs only when ComputeWeight() is called, i.e., 
  at every routing update period.
*/
int IsFeasible( OUTER_CLASS *pClasses, int nClasses, int nLinks,
                int *nBandWidth, int service, int innerclass ) {
    int i, j, k;
    unsigned int allocBandWidth;

    /* add up effective capacities for each link */
    for (i=0;i < nLinks; i++) {
        /* if the link is on the route */
        if (pClasses[service].pInnerClass[innerclass]->nBelong[i]) {
          allocBandWidth = 0;
          for (j=0;j < nClasses;j++) {
            for (k=0;k < pClasses[j].nInnerClass;k++) {
              if (pClasses[j].pInnerClass[k] &&
                  pClasses[j].pInnerClass[k]->nBelong[i]) {
                    allocBandWidth +=
                            pClasses[j].pInnerClass[k]->dN *
                            pClasses[j].pInnerClass[k]->nReserv;
              }
            }
          }
          /* increase by the effective bandwidth that this service requires */ 
          allocBandWidth += pClasses[service].pInnerClass[innerclass]->nReserv;
          if (allocBandWidth > nBandWidth[i]) { /* not feasible */
fprintf(stderr, "Not feasible : link %d, alloc %d\n", nBandWidth[i], allocBandWidth); 
fflush(stderr);
             return 0;  
          }
       }
    }   
    return 1;
}

/* Level-1 feasiblility test. 
   return 1, if the link looks feasible to source,
          0, otherwise

   The model(so does actual network) assumes that throughout an 
  update period, the source has an identical view of the network. 
  Therefore, this runs only when ComputeWeight() is called, i.e., 
  at every routing update period.

   Note that in level1, the feasibility test is based on level-1
  path, i.e., should access interClass.
*/
int IsFeasible2( OUTER_CLASS *pClasses, int nClasses, int nLinks,
                int *nBandWidth, int service, int innerclass ) {
    int i, j, k;
    unsigned int alloc, cap, interclass, zlink, olink, allocBandWidth;

	/* fetch interclass from innerclass */
	interclass = pClasses[service].pInnerClass[innerclass]->intclass_idx;

	/* from level-1 path,
       each iteration takes care of one link and then node
       note that the first element is always level-0 which means we
       don't have to worry about the feasibility on it              */
    for (i=0;i < intclasses[interclass]->numpNodes - 1; i++) {

		if (intclasses[interclass]->pNodes[i].lv == LEVEL0 &&
			intclasses[interclass]->pNodes[i+1].lv == LEVEL0) { 
			
			/* get the link given two endpoints */
			zlink = GetLvZeroLink(intclasses[interclass]->pNodes[i].id,
						  intclasses[interclass]->pNodes[i+1].id);
			zlink--; /* id starts from 1 */

			/* run simliar as is_feasible1() */
            {
              allocBandWidth = 0;
              for (j=0;j < nClasses;j++) {
                for (k=0;k < pClasses[j].nInnerClass;k++) {
                  if (pClasses[j].pInnerClass[k] &&
                      pClasses[j].pInnerClass[k]->nBelong[zlink]) {
                        allocBandWidth +=
                                pClasses[j].pInnerClass[k]->dN *
                                pClasses[j].pInnerClass[k]->nReserv;
                  }
                }
              }
              /* increase by the effective bandwidth that this service requires */ 
              allocBandWidth += pClasses[service].pInnerClass[innerclass]->nReserv;
              if (allocBandWidth > nBandWidth[zlink]) { /* not feasible */
#ifdef DEBUG
    fprintf(stderr, "#Not feasible2 : link %d, alloc %d\n", nBandWidth[zlink], allocBandWidth); 
    fflush(stderr);
#endif
		 f_r++;
                 return 0;  
              } else {
#ifdef DEBUG
    fprintf(stderr, "feasible2 : link %d, alloc %d\n", cap, allocBandWidth);
    fflush(stderr);
#endif
			  }
           }
	
		} else if (intclasses[interclass]->pNodes[i].lv != LEVEL0 ||
			intclasses[interclass]->pNodes[i+1].lv != LEVEL0) {

			/* get the link given on the two endpoints */
    		olink = GetLvOneLink(intclasses[interclass]->pNodes[i].id, 
							   intclasses[interclass]->pNodes[i].lv,
    						   intclasses[interclass]->pNodes[i+1].id, 
							   intclasses[interclass]->pNodes[i+1].lv);

			/* get aggregated acr of the link */
			GetAcrLink(olink, &alloc, &cap);

			/* compare */
			allocBandWidth = alloc + pClasses[service].pInnerClass[innerclass]->nReserv;
			if (allocBandWidth > cap) { /* not feasible */

#ifdef DEBUG
    fprintf(stderr, "#Not feasible2 : link %d, alloc %d\n", cap, allocBandWidth);
    fflush(stderr);
#endif
				f_r++;
				return 0;
			} else {
#ifdef DEBUG
    fprintf(stderr, "feasible2 : link %d, alloc %d\n", cap, allocBandWidth);
    fflush(stderr);
#endif
			}

			/* get aggregated acr of the node */
			if (intclasses[interclass]->pNodes[i+1].lv == LEVEL1) {
				GetAcrNode(intclasses[interclass]->pNodes[i+1].id, &alloc, &cap);

				/* compare */
				allocBandWidth = alloc + pClasses[service].pInnerClass[innerclass]->nReserv;
				if (allocBandWidth > cap) { /* not feasible */

#ifdef DEBUG
    fprintf(stderr, "Not feasible2 : link %d, alloc %d\n", cap, allocBandWidth);
    fflush(stderr);
#endif

					f_r++;
					return 0;
				}
			}
		} else {
			assert(0);
		}
	}
	f_s++;
    return 1;
}

struct tagUpdateStruct {
  OUTER_CLASS* pClasses;
  int nClasses;
  double *dUtilization;
  int nLinks;
};

/*
 * This function updates all the weights of the inner classes.
 *
 * IMPORTANT : add feasibility component here !!!
 *
 */
int ComputeClassWeights( double time, 
                          int nMethod,
                          struct tagUpdateStruct *pUS ) {
  unsigned int nIndex, nInnerIndex, nLinkIndex, nIndex2;

  OUTER_CLASS *pClasses = pUS->pClasses;
  int nClasses = pUS->nClasses;
  double *dUtilization = pUS->dUtilization;
  int nLinks = pUS->nLinks;


  for( nIndex = 0; nIndex < nClasses; nIndex ++ ) { /* foreach service */

    pClasses[ nIndex ].dTotalWeight = 0; 

    for( nInnerIndex = 0; nInnerIndex < pClasses[ nIndex ].nInnerClass ; 
                 nInnerIndex++ ) { /* foreach class */
      
      if( pClasses[ nIndex ].pInnerClass[ nInnerIndex ] != NULL ) {
         /* OldWeight becomes new */
         pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dOldWeight = 
         pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight;

         /* if the method involves utilization in calculation of weights */
        if( nMethod == SEL_UTIL || nMethod == SEL_UTIL_HOP ) {
            double dMaxUtilization = 0.0;

            /* obtain dMaxUtilization */
            for( nLinkIndex = 0; nLinkIndex < nLinks; nLinkIndex ++ ) {
               if( pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->nBelong[ nLinkIndex ] && 
                  dMaxUtilization < dUtilization[ nLinkIndex ] ) {

                  dMaxUtilization = dUtilization[ nLinkIndex ];
               }
            }
            pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight = 
                                1 - dMaxUtilization;
        } else { 
              pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight = 1;
        }

        /* if the method involves hop lengths */
        if( nMethod == SEL_HOP || nMethod == SEL_UTIL_HOP ) {
	   int bidx;

	   bidx = pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->intclass_idx;
           pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight /= 
              (double) (intclasses[ bidx ]->numpNodes - 1);
	   assert(intclasses[ bidx ]->numpNodes > 1);
/*
           pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight /= 
              (double) pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->nHops;
*/
	}

        /* source checks the feasibility here */
        pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight *=
            IsFeasible2(pClasses, nClasses, nLinks, nBandwidth, 
                       nIndex, nInnerIndex);
        
        /* increment the totalWeight of the SERVICE */
        pClasses[ nIndex ].dTotalWeight += 
                    pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight;
      }
    }
    /* At this point, Up and Hp aspects have been accounted */
   
    /* Now, divide all the weights by the total weight of the class */
    for( nInnerIndex = 0; nInnerIndex < pClasses[ nIndex ].nInnerClass ; 
                nInnerIndex++ ) {
      if( pClasses[ nIndex ].pInnerClass[ nInnerIndex ] != NULL ) 
            pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight /=  
                 pClasses[ nIndex ].dTotalWeight;
        /* Also update the lambda in Q  - module */
        pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dL =
pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight * pClasses [ nIndex ].dL;
        Set_Arrival_Rate( pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->nID,
                     pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dL, NULL );

    }
 
    /* And also the z */
    /*
     * #########################################
     *
     *  You may Want to deltet this..
     * 
     * ########################################
     */
    for( nInnerIndex = 0; nInnerIndex < pClasses[ nIndex ].nInnerClass ; 
                nInnerIndex++ ) {
      if( pClasses[ nIndex ].pInnerClass[ nInnerIndex ] != NULL ) {
            for( nIndex2 = 0; nIndex2 < nLinks; nIndex2 ++ ) {

         /* set the oldz and new z to 0 */
              pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dZ[ nIndex2 ] =
              pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dZOld[ nIndex2 ] = 0;

          /*if( pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->nBelong[ nIndex2 ] && 
              pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dOldWeight != 0 ) {
            pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dZ[ nIndex2 ] = 
              ( pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dZ[ nIndex2 ] ) * 
              ( pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight ) / 
              ( pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dOldWeight );
            pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dZOld[ nIndex2 ]= 
              ( pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dZOld[ nIndex2 ] ) * 
              ( pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight ) / 
              ( pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dOldWeight );
            }*/

            }
      }
    }
  } 
}

/* 
 * Just sets the initial Z to 0.. 
 */
void ComputeInitialZ( TRAFFIC_CLASS *classes[], unsigned int nClasses ) {
  int nIndex;
  
  assert( nClasses > 0 );

  while( nClasses -- ) {
    for( nIndex = 0; nIndex < classes[ nClasses ]->nLinks; nIndex ++ )
      classes[ nClasses ]->dZ[ nIndex ] = classes[ nClasses ]->dZOld[ nIndex ] = 0; /*dRatio;  */
  }
}

/*
 * This computes the blocking probability of link 'nLink'
 *
 */
#ifdef GCHITTI
void ComputeBlocking( unsigned int nBandWidth,
                      unsigned int nLink,
                      /* TAKE CARE : Q has size nbandWidth + 1 */
                      TRAFFIC_CLASS* classes[], unsigned int nClasses ) 
{
  int nIndex, nIndexInner;
  double* Q;
  double dSum; /* sum of Q(=link state, reserved bw) values */

  assert( nBandWidth > 0 );
  assert( nClasses > 0 );

  /* allocate memory for link state space */  
  Q = malloc( sizeof( double ) * ( nBandWidth + 1 ));
  if( NULL == Q ) {
    fprintf( stderr," Insufficient Memory! \n " );
    exit( -1 );
  }
  
  Q[ 0 ] = 1;

  dSum = Q[0];
  /* compute all the other Q values, note that they're not probabilities */
  for( nIndex = 1; nIndex <= nBandWidth; nIndex ++ ) { /* for each bw level */
    Q[ nIndex ] = 0;
    /* sum for all relevant classes */
    for( nIndexInner = 0; nIndexInner < nClasses; nIndexInner ++ ) {
      if( nIndex >= classes[ nIndexInner ]->nReserv && 
                  classes[ nIndexInner ]->nBelong[ nLink ] ) {
                 Q[ nIndex ] += ((double) classes[ nIndexInner ]->nReserv) * 
                                classes[ nIndexInner ]->dZ[ nLink ] *
                                Q[ nIndex - classes[ nIndexInner ]->nReserv ];
          }
    }
    Q[ nIndex ] /= nIndex ;  
    dSum += Q[ nIndex ];
  }
  
  /* Compute the blocking probalility */
  for( nIndex = 0; nIndex < nClasses; nIndex ++ ) {
    if( classes[ nIndex ]->nBelong[ nLink ] ) {
      classes[ nIndex ]->dBlocking[ nLink ] = 0;
      /* sum up impossible states */
      for( nIndexInner = 0;
           nIndexInner <  nBandWidth -  classes[ nIndex ]->nReserv + 1;
           nIndexInner ++ ) { 
             classes[ nIndex ]->dBlocking[ nLink ] += Q[ nIndexInner ];
      }
      /* divide by the sum of all Q[] */
      classes[ nIndex ]->dBlocking[ nLink ] = 
                ( dSum - classes[ nIndex ]->dBlocking[ nLink ] ) / dSum;
    }
  }
  
  free( Q );
}
#endif
  
/*
 * This does the iteration for a specific link
 * Computes dZ[], dZOld[] and dBlocking[] 
 */
#ifdef GCHITTI
void DoIteration( unsigned int nBandWidth, unsigned int nLink, 
                          TRAFFIC_CLASS* classes[], unsigned int nClasses, 
                                  double dEpsilon ) 
{

    int nIndex, nInnerIndex;
    int bRecompute;

    do {

      /* set dZ[] from dZOld[] for relevant classes, a little bit akward */
      for( nIndex = 0; nIndex < nClasses; nIndex ++ ) {  
         if( classes[ nIndex ]->nBelong[ nLink ] )
            classes[ nIndex ]->dZ[ nLink ] = classes[ nIndex ]->dZOld[ nLink ]; 
        }

      /* for this link, get the blocking prob. */
      ComputeBlocking( nBandWidth, nLink, classes, nClasses );

      /* Obtain new dZ[] value from dBlocking[] and dNOld[],
       but sotre into dZOld[] */
      for( nIndex = 0; nIndex < nClasses; nIndex ++ ) { 
        if( classes[ nIndex ]->nBelong[ nLink ] ) 
                 classes[ nIndex ]->dZOld[ nLink ] = 
                      ( 1 / ( 1 - classes[ nIndex ]->dBlocking[ nLink ] )) * 
                                  ( classes[ nIndex ]->dNOld ) ;
        }

      /* set the loop flag according to the closeness of dZ[] and dZOld[] */
      bRecompute = 0;
      for( nIndex = 0; nIndex < nClasses && !bRecompute ; nIndex ++ ) {
        if( classes[ nIndex ]->nBelong[ nLink ] ) {
                 double dDif = classes[ nIndex ]->dZ[ nLink ] - 
                                           classes[ nIndex ]->dZOld[ nLink ];
          if( dDif < 0 )
            dDif = -dDif;
          if( dDif > dEpsilon ) 
            bRecompute = 1;
        }
      }

  } while( bRecompute );
}
#endif
         
/*
 * The computation itself : iterates until the given time expires... 
 */
#ifdef GCHITTI
void DoComputation( unsigned int nBandWidth[], double dUtilization[], 
                    unsigned int nLinks, 
                    OUTER_CLASS outerClasses[], unsigned int nOuterClasses,
                    TRAFFIC_CLASS* classes[], unsigned int nClasses, 
                    double dEpsilon, double dDelta, double dInterval, 
                    int nMethod, double dRecomputation,
                    FILE* pfOutFile ) 
{

  int nIndex, nInnerIndex ;
  double dCrtTime, *dSum;
  double dFactorLeft, dFactorRight;
  double dUtil;             /* sum of arrival rates for each class */
  double dBlockPr;          /* average blocking probability */ 
  double dRecomputeTime;    /* upperbound -> dRecomputation */
  int bRecomputedLambda;

  /* sum of interarrival ratio for each service */
  double dLambda = 0;
  for( nIndex = 0; nIndex < nOuterClasses; nIndex ++ ) {
    dLambda +=outerClasses[ nIndex ].dL;
  }

  /* for every class, set their dz[] and dzold[] zero */
  ComputeInitialZ( classes, nClasses );

  /* currently not used */
  dSum = ( double * ) malloc( sizeof( double ) * nLinks );

  /* compute the link utilization and weights */
  /* basically add up effective capacities and divide it by the sum of link
     capacity
  */ 
  if( nMethod == SEL_UTIL || nMethod == SEL_UTIL_HOP ) {
    ComputeUtilization( classes, nClasses, dUtilization, nBandWidth, nLinks );
  }
  ComputeClassWeights( outerClasses, nOuterClasses, dUtilization, nLinks, nMethod, nBandWidth );
   
  dRecomputeTime = 0;
  bRecomputedLambda = 0;

  /* THE MAIN LOOP OF COMPUTATION, at each time step by dDelta */
  for( dCrtTime = 0; dCrtTime < dInterval; dCrtTime += dDelta,
                                           dRecomputeTime+=dDelta ) {

    for( nIndex = 0; nIndex < nClasses; nIndex ++ ) { 
      classes[ nIndex ]->dNOld = classes[ nIndex ]->dN;
    }
    
    /* for each link, obtain dZ[], dZOld[] and dBlocking[] */
    for( nIndex = 0; nIndex < nLinks; nIndex ++ ) {
      DoIteration( nBandWidth[ nIndex ] , nIndex, classes, nClasses, dEpsilon );
    }
    
    /* compute and print the next dN */    
    for( nIndex = 0; nIndex < nLinks; nIndex ++ ) {
      dSum[ nIndex ] = 0;
    }
    dUtil = 0;                   /* sum of the arrival rates of the CLASSES */
    dBlockPr = 0;                /* average blocking probability of CLASSES */
    for( nIndex = 0; nIndex < nClasses; nIndex ++ ) { /* for each class */

      /* Obtain the total blockin probability assuming link independence */
      classes[ nIndex ]->dTotalBlocking = 1.0;
      for( nInnerIndex = 0; nInnerIndex < nLinks ; nInnerIndex ++ ) {
        /* NOTE : actually, this is total NONblocking probability */
        if( classes[ nIndex ]->nBelong[ nInnerIndex ] )
          classes[ nIndex ]->dTotalBlocking *= 
                    1 - classes[ nIndex ]->dBlocking[ nInnerIndex ];
      } 

      /* Set up the equation that obtains new dN from dNOld with 
        service rate, interarrival rate of class and total blocking prob.  
      */
      dFactorLeft =  1 - dDelta / classes[ nIndex ]->dM ;
      if( 0 == classes[ nIndex ]->dLTime ) {
        dFactorRight = classes[ nIndex ]->dL * classes[ nIndex ]->dWeight *
                      classes[ nIndex ]->dTotalBlocking;
      } else {
        dFactorRight = (( int )( dCrtTime / classes[ nIndex ]->dLTime )) 
                        % 2 * classes[ nIndex ]->dL *
                        classes[ nIndex ]->dWeight *
                        classes[ nIndex ]->dTotalBlocking;
      }
      dUtil += dFactorRight; /* increase the total utilization */
      dBlockPr += 1-classes[ nIndex ]->dTotalBlocking; /* increase the total 
                                                blocking probability */
      /* get it ! */
      classes[ nIndex ]->dN = 
              dFactorLeft * classes[ nIndex ]->dNOld + dDelta * dFactorRight;
    }
    dBlockPr /= nClasses;     /* average it */

    /* Is the time to update routing information? */
    /*
    if ( nMethod != SEL_HOP && dRecomputation < dRecomputeTime ) {
    */
    if (dRecomputation < dRecomputeTime ) {

      /* get the new utilization array, necessary? 
         Yes, since it's used by ComputeClassWeights()
      */
      ComputeUtilization( classes, nClasses, dUtilization, nBandWidth, nLinks );
      dRecomputeTime  = 0;

      /* compute the new link weights */
      ComputeClassWeights( outerClasses, nOuterClasses, dUtilization, 
                           nLinks, nMethod, nBandWidth );
      /*
      for( nIndex = 0; nIndex < nClasses; nIndex ++ )
                printf( " ---> %lf\n", classes[ nIndex ]->dWeight );
      */
      
      for( nIndex = 0; nIndex < nLinks; nIndex ++ ) 
                dUtilization[ nIndex ] = 0;
    }
    
    /* print out to pfOutFile */
    fprintf( pfOutFile, "%10.6f ", dCrtTime);
    fprintf( pfOutFile, "%10.6f ", 1-dUtil/dLambda);/* blocking prob. 1 */
    fprintf( pfOutFile, "%10.6f ", dBlockPr);		/* blocking prob. 2 */
    fprintf( pfOutFile, "%10.6f ", avgPathLen());
    fprintf( pfOutFile, "%10.6f ", numCon());

    
        /* calculate the dUtilization[] */
        double *du;
        du = (double *)malloc(sizeof(double) * nLinks);
        assert(du);
        ComputeUtilization( classes, nClasses, du, nBandWidth, nLinks );
        fprintf( pfOutFile, "%10.6f ", avgUtil(du, nLinks));
        free(du);
    
    fprintf( pfOutFile, "%10.6f\n", dUtil);			/* throughput */
	fflush(pfOutFile);

	fprintf(stderr, "dCrtTime = %lf\n", dCrtTime); 
	fflush(stderr);	

  } /* end for */

  fprintf( pfOutFile, "# Feasibility test f_s / (f_s + f_r) = %lf\n", 
                       (double)f_s / (f_s + f_r)); 
  
  free ( dSum );
}
#endif

/* Compute effective bandwidth */
int effect_bw(R, m, b, D, e)
double R, m, b, D, e;
{
   double X, Beff, rho, newBeff, alpha, epsilon;
   int iter;
   epsilon = 0.00001;  /* stopping parameter */
   rho = m / R;
   alpha = b * (1.0 - rho) * R * log( (double) 1.0 / e );
   newBeff = R;
   iter = 0;
   X = 0.0;
 
   do {
      Beff = newBeff;
      if (rho == 1.0) break;
      iter++;
      if (iter > MAX_NUM_ITER) {
        printf("No of iterations for eff bw comp. exceeded %d \n", MAX_NUM_ITER);
        exit(1);
      }
      X = D * Beff;
      newBeff = R * ( alpha - X + sqrt((alpha - X)*(alpha - X) + 4.0*X*rho*alpha) )
                                                      / (2.0 * alpha);
   } while ( fabs((double) (newBeff - Beff)) >  epsilon);
 
   return( (int) ceil(Beff) );
}

int dumper( double crtTime, int classID, double* dN[], double* dBlocking[], 
            double  dAccess, void * pDumpParam) {
  static double dTime;
  static double dTotalN;
  static double dThroughput;
 
  TRAFFIC_CLASS** pClasses = ( TRAFFIC_CLASS ** )pDumpParam;

  /* Another dump cycle */
  if( dTime != crtTime ) {
    /*if( 0.0 != dTime ) */
      printf( "%lf:%lf:%lf\n", dTime, dTotalN, dThroughput );
    dTotalN = 0.0;
    dThroughput = 0.0;
    dTime = crtTime;
  }

  /*printf("Outer Class %d : %.2f +\n",classID ,*dN[0] ); */
  dTotalN += *dN[0];
  dThroughput += pClasses[ classID ]->dL * dAccess;

  return 0;
}


void m_module( double dRecompute, double dInterval )
{
  int i, j, outer_class;
  TRAFFIC_CLASS *cls;
  struct tagUpdateStruct us;
  int callid;

int nclass_at=0;

  us.pClasses = pOuter;
  us.nClasses = nOuter;
  us.dUtilization = dUtilization;
  us.nLinks = nLinks;


  for ( i = 0; i < nLinks; i++) 
    Add_Resource( i, nBandwidth[ i ] );

#ifdef GCHITTI
  printf(" no of links %d \n", nLinks );
  printf(" Link  : bandwidth : Utilization \n" );

  for ( i = 0; i < nLinks; i++) {
    printf("%5d : %5f : %5f \n", glb_pResources[i]->m_nResourceID,
                                 glb_pResources[i]->m_dCapacity,
                                 glb_pResources[i]->m_dUtilization);

  }
#endif

  for ( i = 0; i < nClasses; i++ ) {
    outer_class = classes[i]->serv_idx;
    classes[i]->nID = i;
    /*Add_Class( i , classes[i]->dL / pOuter[ outer_class].nInnerClass );*/
    Add_Class( i , classes[i]->dL );
  } 
#ifdef DEBUG
fprintf(stderr, "nClasses = %d\n", nClasses);
fflush(stderr);
#endif

  for ( i = 0; i < nClasses; i++ ) {
    for ( j = 0; j < nLinks; j++ ) {
       if ( classes[i]->nBelong[j] == 1 ) {
         Add_Class_At_Resource ( i , 
                                 j , 
                                 classes[i]->nReserv ,
                                 1.0/classes[i]->dM );
		nclass_at++;
       }
    }
  }
 
#ifdef DEBUG
fprintf(stderr, "nClass_At = %d\n", nclass_at);
fflush(stderr);
#endif

#ifdef GCHITTI
  printf( " \n no of classes  : %d \n", nClasses );
  printf ( " ClassID : arr rate : serv rate \n");
  for ( i = 0; i < nClasses; i++ )
    printf( " %5d : %5f : %5f \n", glb_pClasses[i]->m_nClassID,
                                   glb_pClasses[i]->m_dArrRate,
                                   glb_pClasses[i]->m_dServRate ); 
#endif

  ComputeClassWeights( 0, nMethod, &us );
  
#ifdef GCHITTI
  printf("\n id  :weights : serv rate \n");
  for ( i = 0; i < nClasses; i++ ) {
    printf(" %i : %5f : %5f : \n", i,
                                       classes[i]->dWeight,
                                       glb_pClasses[i]->m_dServRate ); 
  }
#endif
  
  for ( i = 1; (( i * dRecompute ) < dInterval ); i++ )
    Add_Callback ( ComputeClassWeights, 
                   i * dRecompute, 
                   nMethod, 
                   &us, &callid );

  Register_Dump_Function ( dumper, (void *) classes );


#ifdef GCHITTI
/* want to print all relevant information about the class */
  for ( i = 0; i < nClasses; i++ ) { 
    CLASSES *glb = glb_pClassses[i];
    printf("ID:%d , num_resources %d , Arr %4f , Same:%d ,Serv %4f \n",
             glb->m_nClassID,
             glb->m_nResources ,
              glb->m_dArrRate, 
              glb->m_bSameRate,
              glb->m_dServRate );
    printf ( "Id : blockpr : req'd : servrt : N   : Nold :  Z : Zold \n");
    for ( i = 0; j < glb->m_nResources; j++) {
      printf ("%3d : %7f : %7f : %6f : %5f : %5f : %5f : %5f \n" ,
               glb->m_pResources[j]->m_nResourceID,
               glb->m_pResources[j]->m_nResourceID, );

    }
    

          
  }
#endif


  Evaluate();


  Done_Q_Module();

  /* check clean up */ 
}

/*
 * Infile :
 * <delta>
 * <interval>
 * <epsilon>
 * <recompute time>
 * <method>
 * <nnodes>
 * <nlinks>
 * <link0>
 * <link1>
 * ...
 * <link nlinks>
 * <nOuterClasses>
 * <outer1>
 * <outer2>
 * ...
 * <outer n>
 *
 * <link i> = start stop bandwidth
 * <outer i> = start stop miu lambda bandwidth / 0 Delay epsilonD mc Mc bc  
 *
 * ########### Take care : what os miu in the imput file and here,
 * is actually 1/miu ( normal miu ) ( so it is the mean  service time  ) #######
 */
void main () {
  int nIndex, nInnerIndex, nIndexLink;
  int i;
  struct tagUpdateStruct us;
  int nRuns;

  double dInterval;
  double dDelta;
  double dEpsilon;
  double dRecompute;


  Init_Q_Module();

  scanf( "%d", &nRuns);
  scanf( "%lf", &dDelta ); 
  scanf( "%lf", &dInterval );
  scanf( "%lf", &dEpsilon ); 
  scanf( "%lf", &dRecompute );

  scanf( "%d", &nMethod );
  scanf( "%d", &nNodes );
  scanf( "%d", &nLinks );

  Set_Config_i( CONFIG_NUMBER_OF_SIMULATIONS, nRuns );
  Set_Config_d( CONFIG_EVALUATION_PERIOD, dDelta );
  Set_Config_d( CONFIG_DUMP_PERIOD, dDelta );
 
  Set_Config_d( CONFIG_SIMULATION_TIME, dInterval );

  Set_Config_d( CONFIG_EPSILON_STOP, dEpsilon );

  memset( nLinkMatrix, 0, sizeof( nLinkMatrix ) );

  /***************************************************************/
  /* allocate level-1 id to each level-0 node                    */
  /***************************************************************/
  pNodeGrp = (unsigned int *)malloc(nNodes * sizeof(unsigned int));
  if (pNodeGrp == NULL) {
    fprintf(stderr, "Cannot allocate memory! - abort\n");
    exit(1);
  }
  nLvOneNodes = 0;
  for (i = 0;i < nNodes;i++) {
    int node, grp;
    scanf("%d %d\n", &node, &grp);
    pNodeGrp[node] = grp; 
    
    if (nLvOneNodes < grp)
        nLvOneNodes = grp;
  }
  nLvOneNodes++;    /* note node ids start from 0 */


  /***************************************************************/
  /* assign every level-0 nodes to appropriate grp id - 
     inverse of the above                                        */ 
  /***************************************************************/
  pLvOneNdGrp = (LEVEL_ONE_NODE *)malloc(nNodes * sizeof(LEVEL_ONE_NODE));
  if (pLvOneNdGrp == NULL) {
    fprintf(stderr, "Cannot allocate memory! - abort\n");
    exit(1);
  }
  memset(pLvOneNdGrp, 0, nLvOneNodes * sizeof(LEVEL_ONE_NODE));
  for (i = 0;i < nNodes;i++) {
    pLvOneNdGrp[pNodeGrp[i]].nz[ pLvOneNdGrp[pNodeGrp[i]].num ] = i;
    pLvOneNdGrp[pNodeGrp[i]].num++;
    if (pLvOneNdGrp[pNodeGrp[i]].num == MAX_NODE_GRP) {
        printf("Reached MAX_NODE_GRP = %d : Abort\n", MAX_NODE_GRP); 
        exit(1);
    }
  }

  /***************************************************************/
  /* allocate the stuff for links                                */
  /***************************************************************/
  dUtilization = ( double * ) malloc( nLinks * sizeof( double ));
  memset( dUtilization, 0, sizeof( double ) * nLinks );
  nBandwidth = ( unsigned int * ) malloc( nLinks * sizeof( unsigned int ));
  pLinkEptGrp = (LEVEL_ZERO_LINK *)malloc(nLinks * sizeof(LEVEL_ZERO_LINK));
  
  nLvOneLinks = 0;
  for( nIndex = 0; nIndex < nLinks; nIndex ++ ) { /* for each level-0 link ... */
    int nX, nY;
    /* mark on adjacency matrix */
    scanf( "%d %d %d", &nX, &nY, &nBandwidth[ nIndex ] );
    nLinkMatrix[ nX + nY * MAX_NODES ] = 
        nLinkMatrix[ nY + nX * MAX_NODES ] = nIndex + 1;

    /* assign level-0,1 ids to level-0 links' enpoints */ 
    pLinkEptGrp[nIndex].s = pNodeGrp[nX];
    pLinkEptGrp[nIndex].e = pNodeGrp[nY];
    pLinkEptGrp[nIndex].s0 = nX;
    pLinkEptGrp[nIndex].e0 = nY;

    /* aggregate level-1 links */
    if (pNodeGrp[nX] != pNodeGrp[nY]) {/* if in the same grp,don't aggregate */
        unsigned int idx;

        /* pNodeGrp[nX].nX, pNodeGrp[nY].nY :
           1. (pNodeGrp[nX], pNodeGrp[nY]) 
           2. (pNodeGrp[nX].nX, pNodeGrp[nY])
           3. (pNodeGrp[nY].nY, pNodeGrp[nX])
        */
        /* 1. record in pLvOneLkGrp[idx] : num++, type:0, name, link:nIndex */  
        idx = LkOneTypeZeroLookUp(pNodeGrp[nX], pNodeGrp[nY]);
        pLvOneLkGrp[idx].type = LINK_AGR_TYPE0;
        pLvOneLkGrp[idx].id.s0 = min(pNodeGrp[nX], pNodeGrp[nY]);
        pLvOneLkGrp[idx].id.e = max(pNodeGrp[nX], pNodeGrp[nY]);
        pLvOneLkGrp[idx].nz[pLvOneLkGrp[idx].num++] = nIndex+1;
        if (pLvOneLkGrp[idx].num == MAX_LINK_GRP) {
            printf("Reached MAX_LINK_GRP = %d : Abort\n", MAX_LINK_GRP); 
            exit(1);
        }

        /* 2. record in pLvOneLkGrp[idx] : num++, type:1, name, link:nIndex */
        idx = LkOneTypeOneLookUp(pNodeGrp[nX], nX, pNodeGrp[nY]);
        pLvOneLkGrp[idx].type = LINK_AGR_TYPE1;
        pLvOneLkGrp[idx].id.s0 = pNodeGrp[nX];
        pLvOneLkGrp[idx].id.s1 = nX;
        pLvOneLkGrp[idx].id.e = pNodeGrp[nY];
        pLvOneLkGrp[idx].nz[pLvOneLkGrp[idx].num++] = nIndex+1;
        if (pLvOneLkGrp[idx].num == MAX_LINK_GRP) {
            printf("Reached MAX_LINK_GRP = %d : Abort\n", MAX_LINK_GRP); 
            exit(1);
        }

        /* 3. record in pLvOneLkGrp[idx] : num++, type:1, name, link:nIndex */
        idx = LkOneTypeOneLookUp(pNodeGrp[nY], nY, pNodeGrp[nX]);
        pLvOneLkGrp[idx].type = LINK_AGR_TYPE1;
        pLvOneLkGrp[idx].id.s0 = pNodeGrp[nY];
        pLvOneLkGrp[idx].id.s1 = nY;
        pLvOneLkGrp[idx].id.e = pNodeGrp[nX];
        pLvOneLkGrp[idx].nz[pLvOneLkGrp[idx].num++] = nIndex+1;
        if (pLvOneLkGrp[idx].num == MAX_LINK_GRP) {
            printf("Reached MAX_LINK_GRP = %d : Abort\n", MAX_LINK_GRP); 
            exit(1);
        }

    } 
  }

  /***************************************************************/
  /* Build views for each level-1 node                           */
  /***************************************************************/
  view = (VIEW *)malloc(nLvOneNodes * sizeof(VIEW));
  if (view == NULL) {
     fprintf(stderr, "Cannot allocate memory! - abort\n");
     exit(1);
  }
  memset(view, 0, nLvOneNodes * sizeof(VIEW));

  for (i=0;i < nLvOneNodes;i++) { /* for each view */
     int j,k;

     /* NODES */
     /* number of level-0 nodes */
     view[i].nLvZeroNodes = view[i].nNodes = pLvOneNdGrp[i].num;

     /* number of level-1 nodes */
     view[i].nNodes += (nLvOneNodes-1);

     /* construct the node list */
     view[i].pNd = (struct tag_vn *)malloc(view[i].nNodes * 
                                           sizeof(struct tag_vn));
     if (view[i].pNd == NULL) {
         fprintf(stderr, "Cannot allocate memory! - abort\n");
         exit(1);
     }
     for (j=0;j < pLvOneNdGrp[i].num;j++) {
        view[i].pNd[j].id = pLvOneNdGrp[i].nz[j];  
        view[i].pNd[j].lv = 0;
     }
     for (j=0,k=view[i].nLvZeroNodes;j < nLvOneNodes;j++) {
        if (i != j) { 
            view[i].pNd[k].id = j;  
            view[i].pNd[k++].lv = 1;
        }
     }

     /* LINKS */
     /* allocate adjacency matrix */
     view[i].pAj = malloc(view[i].nNodes * view[i].nNodes * 
                          sizeof(struct tag_vl));
     if (view[i].pAj == NULL) {
         fprintf(stderr, "Cannot allocate memory! - abort\n");
         exit(1);
     }
     memset(view[i].pAj, 0, view[i].nNodes * view[i].nNodes * 
                                             sizeof(struct tag_vl));

     /* mark level-0 links */
     for (j=0;j < nLinks;j++) {
        if (pLinkEptGrp[j].s == i && pLinkEptGrp[j].e == i) {
            unsigned int start, end;

            start = GetNdIdx(LEVEL0, pLinkEptGrp[j].s0, &view[i]); 
            end = GetNdIdx(LEVEL0, pLinkEptGrp[j].e0, &view[i]); 
            /* link ids start from 1 */
            view[i].pAj[start + end * view[i].nNodes].id = j+1;
            view[i].pAj[start + end * view[i].nNodes].lv = 0;
            view[i].pAj[end + start * view[i].nNodes].id = j+1;
            view[i].pAj[end + start * view[i].nNodes].lv = 0;
        }
     }
     
     /* mark level-1 links */
     for (j=0;j < nLvOneLinks;j++) { /* for each level-1 link */
        if (pLvOneLkGrp[j].type == LINK_AGR_TYPE1) { /* (x.i, y) */
            if (pLvOneLkGrp[j].id.s0 == i) {
              unsigned int start, end;

              start = GetNdIdx(LEVEL0, pLvOneLkGrp[j].id.s1, &view[i]);
              end = GetNdIdx(LEVEL1, pLvOneLkGrp[j].id.e, &view[i]);
              /* link ids start from 1 */
              view[i].pAj[start + end * view[i].nNodes].id = j+1;
              view[i].pAj[start + end * view[i].nNodes].lv = 1;
              view[i].pAj[end + start * view[i].nNodes].id = j+1;
              view[i].pAj[end + start * view[i].nNodes].lv = 1;
            }
        } else { /* LINK_AGR_TYPE0 , (y,z) */
            if (pLvOneLkGrp[j].id.s0 != i && pLvOneLkGrp[j].id.e != i) {
              unsigned int start, end;

              start = GetNdIdx(LEVEL1, pLvOneLkGrp[j].id.s0, &view[i]);
              end = GetNdIdx(LEVEL1, pLvOneLkGrp[j].id.e, &view[i]);
              /* link ids start from 1 */
              view[i].pAj[start + end * view[i].nNodes].id = j+1;
              view[i].pAj[start + end * view[i].nNodes].lv = 1;
              view[i].pAj[end + start * view[i].nNodes].id = j+1;
              view[i].pAj[end + start * view[i].nNodes].lv = 1;
            }
        }
     }
     
  }

  
  /***************************************************************/
  /* process services                                            */
  /***************************************************************/

  /* read in number of services */
  scanf( "%d", &nOuter );
  pOuter = ( OUTER_CLASS* ) malloc( sizeof( OUTER_CLASS ) * nOuter );
  if( NULL == pOuter ) {
    free( nBandwidth );
    exit( -1 );
  }
  
  for( nIndex = 0; nIndex < nOuter; nIndex ++ ) { /* for each service .... */
    scanf( "%d %d", &pOuter[ nIndex ].nStartNode, &pOuter[ nIndex ].nStopNode );
    scanf( "%d", &pOuter[ nIndex ].nReserv );

    /* read lambda / miu */
    scanf( "%lf %lf ", &pOuter[ nIndex ].dL, &pOuter[ nIndex ].dM );
    pOuter[ nIndex ].dLTime = 0.0;

    /* lambda in steps ? */
    if( 0 == pOuter[ nIndex ].dL ) 
      scanf( "%lf %lf ", &pOuter[ nIndex ].dL, &pOuter[ nIndex ].dLTime );

    /* the reserved bandwidth .. */
    if( 0 == pOuter[ nIndex ].nReserv ) {
      /* special type .. */
      float dDelay, dEpsilonD, dmc, dMc, dbc;
      double dBeta, dXjc, dXjcOld;
      double dReserv;
      scanf( "%f %f %f %f %f", &dDelay, &dEpsilonD, &dmc, &dMc, &dbc );
     
      /* compute the bandwidth ! */
      dBeta = log( 1 / dEpsilonD ) * dbc * ( dMc - dmc );
      dXjc = 0; 

      do {
        dReserv = dMc * ( dBeta - dXjc + 
                          sqrt(( dBeta - dXjc ) * ( dBeta - dXjc ) + 
                               4.0 * dXjc * dmc * dBeta / dMc ))/2.0/dBeta;
        dXjcOld =  dXjc;
        dXjc = dReserv * dDelay;
      } while( dXjc -  dXjcOld > 0 ? dXjc -  dXjcOld : dXjcOld -  dXjc > dDelay );

      /* save it ! */
      pOuter[ nIndex ].nReserv = ceil( dReserv );
    }

    pOuter[ nIndex ].nInnerClass = 0;
    pOuter[ nIndex ].pInnerClass = NULL;
    pOuter[ nIndex ].nInterClass = 0;
    pOuter[ nIndex ].pInterClass = NULL;
    pOuter[ nIndex ].nMinHops = nNodes;
    pOuter[ nIndex ].nLvOneMinHops = nNodes;
  }
  /***************************************************************/
  /* compute the level-1 routes                                  */
  /***************************************************************/
  nIntClasses = 0;
  ComputeLvOneRoutes();

/*
PrintLevelZeroGrpId();
PrintLevelZeroLinkEndPtr();
PrintView();
PrintInterClass(1);
PrintServices(pOuter, nOuter);
exit(0);
*/

  /***************************************************************/
  /* estimate the routing update cost (static)					 */ 
  /***************************************************************/
#ifdef STATIC_EVAL
  {
	unsigned int tmp;
    tmp = getLvZeroRTSize();
    printf("# Level-0 RT size = %d\n", tmp);
    tmp = getLvOneRTSize();
    printf("# Level-1 RT size = %d\n", tmp);
    tmp = getLvZeroComCost();
    printf("# Level-0 COM cost = %d\n", tmp);
    tmp = getLvOneComCost();
    printf("# Level-1 COM cost = %d\n", tmp);
  }
#endif
  
  /***************************************************************/
  /* compute the level-0 routes                                  */
  /***************************************************************/

  BuildClasses();

  us.pClasses = pOuter;
  us.nClasses = nOuter;
  us.dUtilization = dUtilization;
  us.nLinks = nLinks;


  /***************************************************************/
  /* main computation											 */
  /***************************************************************/
  /* nBandwidth - each link's capacity, (static)
     dUtilization - each link's utilization,
     dRecompute - update period 
  */
#ifdef GCHITTI
  DoComputation( nBandwidth, dUtilization, nLinks, 
                 pOuter, nOuter, 
                 classes, nClasses, 
                 dEpsilon, dDelta, dInterval, nMethod, dRecompute, stdout );
#endif

  m_module(dRecompute, dInterval);

  for( nIndex = 0; nIndex < nClasses; nIndex ++ ) {
    free( classes[ nIndex ]->nBelong ); 
    free( classes[ nIndex ]->dBlocking );
    free( classes[ nIndex ]->dZ );
    free( classes[ nIndex ]->dZOld );
    free( classes[ nIndex ] );
  }

  for( nIndex = 0; nIndex < nOuter; nIndex ++ ) 
    free( pOuter[ nIndex ].pInnerClass );

  free( classes );
  free( nBandwidth );
  free( pOuter );
  free( dUtilization );
}
