/*
 * This is an example of M - Module. It computes the parameters
 * for a network providing real-time services. 
 *
 * Catalin Popescu    Sept 23 '97
 */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <memory.h>
#include <math.h>

#include "mcmr.h"
#define TRUE 1
#define FALSE 0

#define SEL_HOP       1
#define SEL_UTIL      2
#define SEL_UTIL_HOP  3
#define MAX_NODES     50

/***********************************************
 * 
 *  Local Structures
 *
 */


/* 
 * The Inner Class Structure
 */
typedef struct tagTRAFFIC_CLASS {
  unsigned int nID;
  /* The id of the class */

  double dL;                
  /* arrival rate */
  double dM; 
  /* serving rate */

  double dWeight;
  /* The weight of this inner class between the other inner
   * classe of the same outer class
   */
  double dOldWeight;
  /* The old weight */

  unsigned int nReserv; 
  /* required Bandwidth */

  unsigned int nHops;
  /* The length of the path of the class */

  unsigned int *nBelong;
  /* A vector with dimmension nLinks, to hold the class's path */
} TRAFFIC_CLASS;

/*
 * This structure holds an outer class information
 */
typedef struct tagOUTER_CLASS {

  double dL; 
  /* arrival rate */
  double dM; 
  /* serving rate */

  double dTotalWeight;
  /* The sum of the weights of bellonging inner classes */

  /* From where starts and where ends the connection */
  unsigned int nStartNode;
  unsigned int nStopNode;

  /* The bandwidth requirement */
  unsigned int nReserv;

  /* the number of inner classes */
  unsigned int nInnerClass;
  /* The inner classes */
  TRAFFIC_CLASS** pInnerClass;

  /* The minimum number of hops of inner classes */
  unsigned int nMinHops;

} OUTER_CLASS;


/*
 * This funtion, is recursive one, and will compute all the routes from nNodes[0] to 
 * nStopNode. The routes will be stored in the pClass traffic class.
 *
 * Input :
 *   int pMatrix[] - the topology of the network
 *   OUTER_CLASS* pClass - the class we want to build the routes
 *   unsigned int nNodes - the number of nodes in the network
 *   unsigned int nCrtNode - the numbers of nodes in path till now.
 *   int pVisited[] - an array to prevent circular paths
 *   unsigned int nNodes[] - the path till now.
 *   unsigned int nLinks - the total number of links in the network.
 */
void ComposeRoute( int pMatrix[], OUTER_CLASS* pClass, 
		   unsigned int nNodes, unsigned int nCrtNode,
		   int pVisited[], unsigned int pNodes[], unsigned int nLinks
		  ) {  
  int nIndex;

  /* Already too far..  discard this path */
  if( nCrtNode  > pClass->nMinHops + 1  )
    return;

  if( pClass->nStopNode == pNodes[ nCrtNode ] ) {
    TRAFFIC_CLASS *pNewClass;
    TRAFFIC_CLASS **pNewClasses = ( TRAFFIC_CLASS ** ) malloc( sizeof( TRAFFIC_CLASS * ) * ( pClass->nInnerClass + 1 ));

    if( pClass->nInnerClass > 0 ) {
      /* if is not the first route, copy the other routes */
      assert( pClass->pInnerClass != NULL );
      memcpy( pNewClasses, pClass->pInnerClass, pClass->nInnerClass * sizeof( TRAFFIC_CLASS * ) );
      free( pClass->pInnerClass );
    }
    
    /* Craete and initialize the new Class */
    pNewClass = ( TRAFFIC_CLASS * ) malloc( sizeof( TRAFFIC_CLASS ));

    pNewClass->dL = pClass->dL;
    pNewClass->dM = pClass->dM;
    
    pNewClass->nReserv = pClass->nReserv;
    pNewClass->dWeight = pNewClass->dOldWeight = 0;
    pNewClass->nBelong = ( unsigned int *) malloc( sizeof( unsigned int ) * nLinks );
    memset( pNewClass->nBelong, 0, sizeof( unsigned int ) * nLinks );

    pNewClass->nHops = nCrtNode;

    for( nIndex = 1; nIndex <= nCrtNode; nIndex ++ )
      pNewClass->nBelong[ pMatrix[ pNodes[ nIndex - 1 ] +  pNodes[ nIndex ] * MAX_NODES ] - 1 ]  = 1;  
 
    pNewClasses[ pClass->nInnerClass ] = pNewClass;

    /* Update the hops */
    if( nCrtNode  < pClass->nMinHops )
      pClass->nMinHops = nCrtNode;
    
    pClass->nInnerClass ++;
    pClass->pInnerClass = pNewClasses;
  }
  else {
    for( nIndex = 0; nIndex < nNodes; nIndex ++ ) {
      /* A new possible path.. ? */
      if( pMatrix[ pNodes[ nCrtNode ] + nIndex * MAX_NODES ] != 0 && !pVisited[ nIndex ] ){
	/* got it ..! */
	pVisited[ nIndex ] = TRUE;
	pNodes[ nCrtNode + 1 ] = nIndex;
	
	/* next step */
	ComposeRoute( pMatrix, pClass, nNodes, nCrtNode + 1, 
		      pVisited, pNodes, nLinks );
	
	/* rollback */
	pVisited[ nIndex ] = FALSE;
      }
    }
  }
}

/*
 * This function will compute for each class all possible routes. Usses the ComposeRoute
 * function to achieve this.
 *
 * Input :
 *   int pMartix[] - the adjacency matrix
 *   unsigned int nNodes - the number of nodes in the network
 *   OUTER_CLASS pClasses[] - the outer classes
 *   unsigned int nClasses - the number of outer Classes
 *   unsigned int nLinks - the number of edges in the network's graph
 *   
 * Output :
 *   TRAFFIC_CLASS** pReturnClasses - the inner classes computed in this function
 *   int* pNumClasses - the number of inner classes
 * 
 */
void ComputeClassRoutes( int pMatrix[], unsigned int nNodes,
			 OUTER_CLASS pClasses[], unsigned int nClasses,
			 unsigned int nLinks,
			 TRAFFIC_CLASS *** pReturnClasses, int* pNumClasses ) {
  unsigned int nIndex, nInnerIndex, nIndexClass;
  unsigned int* pNodes;
  int* pVisited;
  unsigned int nInnerClasses = 0;

  /* Allocate some buffers */
  pNodes = ( unsigned int* ) malloc( nNodes * sizeof( unsigned int ));
  if( NULL == pNodes  ) {
    fprintf( stderr, "Error: Out of memory.. \n");
    exit( -1 );
  }
  pVisited = ( int* ) malloc( nNodes * sizeof( int ));
  if( NULL == pVisited ) {
    fprintf( stderr, "Error: Out of memory.. \n");
    exit( -1 );
  }

  /* compute routes for each class */
  for( nIndex = 0; nIndex < nClasses; nIndex++ ) {
    /* nothing visited yet. */
    memset( pVisited, 0, nNodes * sizeof( int ));

    for( nInnerIndex = 0; nInnerIndex < nNodes; nInnerIndex ++ )
      pNodes[ nInnerIndex ] = MAX_NODES+1;

    pClasses[nIndex].nInnerClass = 0;
    pClasses[nIndex].pInnerClass = NULL;
    pClasses[nIndex].nMinHops = nNodes;

    /* start the route */
    pNodes[0] = pClasses[nIndex].nStartNode;
    pVisited[pClasses[nIndex].nStartNode] = TRUE;
    if( pClasses[ nIndex ].nStartNode == pClasses[ nIndex ].nStopNode )
      printf( "ERROR - class %d from %d to %d !\n", nIndex, pClasses[ nIndex ].nStartNode,
	      pClasses[ nIndex ].nStopNode );

    /* do the job ! */
    ComposeRoute( pMatrix, &pClasses[nIndex], nNodes, 0, pVisited, pNodes,
		  nLinks );

    /* now, reduce the routes to the effective ones.. */
    for( nInnerIndex = 0; nInnerIndex < pClasses[nIndex].nInnerClass; nInnerIndex ++ ) 
      if( pClasses[nIndex].pInnerClass[nInnerIndex]->nHops > pClasses[nIndex].nMinHops + 1 ) {
	/* free the inner class !! */
	free( pClasses[nIndex].pInnerClass[nInnerIndex]->nBelong );
	
	free(  pClasses[nIndex].pInnerClass[nInnerIndex] );
	pClasses[nIndex].pInnerClass[nInnerIndex] = NULL;
      }
      else
	(*pNumClasses) ++;
  }	
  /* Now, just try to save the data .. */
  (*pReturnClasses) = ( TRAFFIC_CLASS ** ) malloc( sizeof( TRAFFIC_CLASS * ) * *pNumClasses );

  nIndexClass = 0;
  for( nIndex = 0; nIndex < nClasses; nIndex ++ )
    for( nInnerIndex = 0; nInnerIndex < pClasses[nIndex].nInnerClass; nInnerIndex ++ ) 
      if( pClasses[nIndex].pInnerClass[nInnerIndex] != NULL ) {
	(*pReturnClasses)[ nIndexClass ] = pClasses[nIndex].pInnerClass[nInnerIndex];
	nIndexClass ++;
      }  

  /* cleanup.. */
  free( pVisited );
  free( pNodes );
}

/*
 * This function updates all the weights of the inner classes.
 */
struct tagUpdateStruct {
  OUTER_CLASS* pClasses;
  int nClasses;
  double *dUtilization;
  int nLinks;
};
/*
 * void ComputeClassWeights( double time, int nMethod, struct tagUpdateStruct* pUS )
 * 
 * This is a callback function to update the routing. It has as parameters :
 *
 * Input :
 *    double time - the current time of evaluation
 *    int nMethod - the method to update routes. One of :
 *                        SEL_HOP, SEL_UTIL or SEL_UTIL_HOP
 *    struct tagUpdateStruct *pUS - this is pParam, and contains information about the
 *                        structure of the classes.
 */
int ComputeClassWeights( double time, int nMethod, struct tagUpdateStruct* pUS ) {  
  OUTER_CLASS *pClasses;
  int nClasses;
  double *dUtilization;
  int nLinks;

  unsigned int nIndex, nInnerIndex, nIndex2;

  /* Fill in the local variables from the structure */
  pClasses = pUS->pClasses;
  nClasses = pUS->nClasses;
  dUtilization = pUS->dUtilization;
  nLinks = pUS->nLinks;

  /* Now, if it is not the time 0, ask Q- module to recompute the utilization */
  if( time > 0 ) {
    Recompute_Resource_Utilization();
    /* and collect the results .. */
    for( nIndex = 0; nIndex < nLinks; nIndex ++ ) 
      Fetch_Crt_Basic_Metrics( METRIC_UTILIZATION, 0, nIndex, 
							   &dUtilization[ nIndex ] );
  }

  /* Now we are ready.. !!! */
  for( nIndex = 0; nIndex < nClasses; nIndex ++ ) {
    pClasses[ nIndex ].dTotalWeight = 0; 
    for( nInnerIndex = 0; nInnerIndex < pClasses[ nIndex ].nInnerClass ; nInnerIndex++ ) {
      if( pClasses[ nIndex ].pInnerClass[ nInnerIndex ] != NULL ) {
	pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dOldWeight = 
	  pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight;
	
	/* Different method if we have to use utilization or not */
	if( nMethod == SEL_UTIL || nMethod == SEL_UTIL_HOP ) {
	  int nLinkIndex;
	  double dMaxUtilization = 0.0;

	  /* Get the maximum utilization by browsing through links */
	  for( nLinkIndex = 0; nLinkIndex < nLinks; nLinkIndex ++ )
	    if( pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->nBelong[ nLinkIndex ] && 
		dMaxUtilization < dUtilization[ nLinkIndex ] )
	      dMaxUtilization = dUtilization[ nLinkIndex ];

	  /* Got it ! - Update it ! */
	  pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight = 1 - dMaxUtilization;
	}

	else 
	  /* only hops - no utilization */
	  pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight = 1;

	/* Now, if we have to use also hops - use them .. */
	if( nMethod == SEL_HOP || nMethod == SEL_UTIL_HOP )
	  pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight /= 
	    (double) pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->nHops;
	
	/* increment the totalWeight */
	pClasses[ nIndex ].dTotalWeight += pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight;
      }
    }
   
    /* Now, update all the weights */
    for( nInnerIndex = 0; nInnerIndex < pClasses[ nIndex ].nInnerClass ; nInnerIndex++ )
      if( pClasses[ nIndex ].pInnerClass[ nInnerIndex ] != NULL ) {
	pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight /=  pClasses[ nIndex ].dTotalWeight;
	/* Also update the lambda in Q  - module */
	pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dL =
	  pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dWeight * pClasses[ nIndex ].dL;
	Set_Arrival_Rate( pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->nID, 
			  pClasses[ nIndex ].pInnerClass[ nInnerIndex ]->dL, NULL );
      }
  } 
}

/*
 * This implements a numerical method to determine the required 
 * ammount of bandwith givven some other parameters..
 *
 * Nothing spectacular
 */
void ComputeReserved( OUTER_CLASS* pOuter ) {
  /* special type .. */
  float dDelay, dEpsilonD, dmc, dMc, dbc;
  double dBeta, dXjc, dXjcOld;
  double dReserv;
  scanf( "%f %f %f %f %f", &dDelay, &dEpsilonD, &dmc, &dMc, &dbc );
  
  /* compute the bandwidth ! */
  dBeta = log( 1 / dEpsilonD ) * dbc * ( dMc - dmc );
  dXjc = 0; 
  
  do {
    dReserv = dMc * ( dBeta - dXjc + 
		      sqrt(( dBeta - dXjc ) * ( dBeta - dXjc ) + 
			   4.0 * dXjc * dmc * dBeta / dMc ))/2.0/dBeta;
    dXjcOld =  dXjc;
    dXjc = dReserv * dDelay;
  } while( dXjc -  dXjcOld > 0 ? dXjc -  dXjcOld : dXjcOld -  dXjc > dDelay );

  /* save it ! */
  pOuter->nReserv = ceil( dReserv );
}

/*
 * This is the Dump function - see Q - module for details. It dumps now only N..
 *
 * Can be improved, of cpourse.. Depends what you want .. 
 */
int dumper( double crtTime, int classID, double* dN[], double* dBlocking[], double  dAccess,
	    void * pDumpParam) {
  static double dTime;
  static double dTotalN;
  static double dThroughput;
  
  TRAFFIC_CLASS** pClasses = ( TRAFFIC_CLASS ** )pDumpParam;

  /* Another dump cycle */
  if( dTime != crtTime ) {
    if( 0.0 != dTime ) 
      printf( "%lf:%lf:%lf\n", dTime, dTotalN, dThroughput );
    dTotalN = 0.0;
    dThroughput = 0.0;
    dTime = crtTime;
  }

  dTotalN += *dN[0];
  dThroughput += pClasses[ classID ]->dL * dAccess;

  return 0;
}

static int nLinkMatrix[ MAX_NODES * MAX_NODES ];

/*
 * The main Function - mother of all ...
 *
 * It reads first the input, creates the classes and ressources, registers the callbacks and 
 * passes control to Q - module. After this - does the cleanup .
 */
void main () {
  int nIndex, nInnerIndex, nIndexLink, mIndex;

  /* inner classes */
  unsigned int nClasses;
  TRAFFIC_CLASS **classes;

  /* outer classes */
  unsigned int nOuter;
  OUTER_CLASS *pOuter;

  /* The parameters of the links */
  unsigned int  nLinks;
  unsigned int *nBandwidth;
  double       *dUtilization;

  /* Something to be read from the input file */
  double dInterval, dDelta, dEpsilon, dRecompute;

  unsigned int nNodes;
  int          nMethod;
  int          id, nRuns;

  /* A structure used to pass parameters to weight update function */
  struct tagUpdateStruct us;

  /* 
   * Init the Q - module 
   */
  Init_Q_Module();

  /* 
   * Now read som variables, and set them in Q - module 
   */
  scanf( "%d", &nRuns);
  Set_Config_i( CONFIG_NUMBER_OF_SIMULATIONS, nRuns );
  scanf( "%lf", &dDelta ); 
  Set_Config_d( CONFIG_EVALUATION_PERIOD, dDelta );
  Set_Config_d( CONFIG_DUMP_PERIOD, dDelta );
  
  scanf( "%lf", &dInterval );
  Set_Config_d( CONFIG_SIMULATION_TIME, dInterval );

  scanf( "%lf", &dEpsilon ); 
  Set_Config_d( CONFIG_EPSILON_STOP, dEpsilon );

  /* This holds the link update period.. */
  scanf( "%lf", &dRecompute );
  
  /* Method for routing */
  scanf( "%d", &nMethod );

  /* # of links and nodes */
  scanf( "%d", &nNodes );
  scanf( "%d", &nLinks );

  /* 
   * Allocate the adjacency matrix
   */
  memset( nLinkMatrix, 0, sizeof( nLinkMatrix ) );

  /* 
   * allocate the stuff for links 
   */
  nBandwidth   = ( unsigned int * ) malloc( nLinks * sizeof( unsigned int ));
  dUtilization = ( double * )       malloc( nLinks * sizeof( double ));
  memset( dUtilization, 0, sizeof( double ) * nLinks );

  /* 
   * Read the edges
   */
  for( nIndex = 0; nIndex < nLinks; nIndex ++ ){
    int nX, nY;
    scanf( "%d %d %d", &nX, &nY, &nBandwidth[ nIndex ] );
    /* Update the matrix */
    nLinkMatrix[ nX + nY * MAX_NODES ] = nLinkMatrix[ nY + nX * MAX_NODES ] = nIndex + 1;
  }

  /*
   * Read the # of classes and allocate the vector to hold them 
   */
  scanf( "%d", &nOuter );
  pOuter = ( OUTER_CLASS* ) malloc( sizeof( OUTER_CLASS ) * nOuter );
  if( NULL == pOuter ) {
    /* Error... Cleanup and exit */
    Done_Q_Module();
    free( nBandwidth );
    free( dUtilization );
    exit( -1 );
  }
  
  /* 
   * Read the classes 
   */
  for( nIndex = 0; nIndex < nOuter; nIndex ++ ) {
    scanf( "%d %d", &pOuter[ nIndex ].nStartNode, &pOuter[ nIndex ].nStopNode );
    scanf( "%d", &pOuter[ nIndex ].nReserv );

    /* read lambda / miu */
    scanf( "%lf %lf ", &pOuter[ nIndex ].dL, &pOuter[ nIndex ].dM );

    /********* Take Care - read time - convert to rate  ********/
    pOuter[ nIndex ].dM = 1.0 / pOuter[ nIndex ].dM;

    /* Compuute the reserved bandwidth .. */
    if( 0 == pOuter[ nIndex ].nReserv ) 
      ComputeReserved( &pOuter[ nIndex ] );
  }

  /* fill in the route update structure */
  us.pClasses = pOuter;
  us.nClasses = nOuter;
  us.dUtilization = dUtilization;
  us.nLinks = nLinks;
  
  /* compute the routes */
  nClasses = 0;
  ComputeClassRoutes( nLinkMatrix, nNodes, pOuter, nOuter, nLinks, &classes, &nClasses );

  /* now, let's add everything .. No much checking..*/
  for( nIndex = 0; nIndex < nLinks; nIndex ++ )
    Add_Resource( nIndex, nBandwidth[ nIndex ] );
  
  for( nIndex = 0; nIndex < nClasses; nIndex ++ ) {
    /* Add the class */
    classes[ nIndex ]->nID = nIndex;
    Add_Class( nIndex, classes[ nIndex ]->dL );
  }
  for( nIndex = 0; nIndex < nClasses; nIndex ++ ) {
    /* Set it at the ressource */
    for( mIndex = 0; mIndex < nLinks; mIndex ++  ) {
      if( classes[ nIndex ]->nBelong[ mIndex ] )
	Add_Class_At_Resource( nIndex, mIndex, 
			       classes[ nIndex ]->nReserv, classes[ nIndex ]->dM );
    }
  }

  /* Update the weights */
  ComputeClassWeights( 0, nMethod, &us );

  /*
   * Now, register the callbacks for routing update 
   */  
  for( nIndex = 1; (( nIndex * dRecompute ) < dInterval ); nIndex ++ )  
    Add_Callback( ComputeClassWeights, nIndex * dRecompute, nMethod, 
					(void*)&us, &id );

  /*
   * I also want to register a dump function
   */
  Register_Dump_Function( dumper, ( void* )classes );

  /*
   * Start the evaluation
   */
  Evaluate ();

  /*
   * Finish it !
   */
  Done_Q_Module();

  /*
   * The big cleanup 
   */
  for( nIndex = 0; nIndex < nClasses; nIndex ++ ) {
    free( classes[ nIndex ]->nBelong ); 
    free( classes[ nIndex ] );
  }

  for( nIndex = 0; nIndex < nOuter; nIndex ++ ) 
    free( pOuter[ nIndex ].pInnerClass );

  free( classes );
  free( nBandwidth );
  free( pOuter );
  free( dUtilization );
}
