/*
 * The main q-z module implementation
 */

#include <stdlib.h>
#include <errno.h>
#include <assert.h>

#include "mcmr.h"
#include "structures.h"
#include "zit.h"


/*
 * the registered events / callbacks 
 */
EVENT** glb_pEventList;
unsigned int glb_nEvents;
unsigned int glb_nEventID;
/*
 * Global classes and ressources
 */
unsigned int glb_nClasses;
CLASS ** glb_pClasses;

unsigned int glb_nResources;
RESOURCE ** glb_pResources;

/* cross referrence matrix for the Z and Blocking measures */
double * glb_pZ;
double * glb_pZOld;
double * glb_pBlocking;
double * glb_pN;
double * glb_pNOld;
double * glb_pRequired;
double * glb_pServRate;
unsigned int glb_nAssoc;

/* 
 * Some get ressource / class functions
 */
CLASS* getClassByID( unsigned int nID ) {
  unsigned int nIndex;

  for( nIndex = 0; nIndex < glb_nClasses; nIndex ++ ) {

    if( glb_pClasses[ nIndex ] != NULL )
      if( glb_pClasses[ nIndex ]->m_nClassID == nID )
	return glb_pClasses[ nIndex ] ;
  }

  return NULL;
}

RESOURCE* getResourceByID( unsigned int nID ) {
  unsigned int nIndex;

  for( nIndex = 0; nIndex < glb_nResources; nIndex ++ ) {

    if( glb_pResources[ nIndex ] != NULL )
      if( glb_pResources[ nIndex ]->m_nResourceID == nID )
	return glb_pResources[ nIndex ]; 
  }

  return NULL;
}

/*
 * Configuration settings 
 */
unsigned int glb_nArrivalDistribution;
unsigned int glb_nServiceDistribution;

double glb_dSimulationTime;
double glb_dDumpPeriod;
double glb_dEvaluationPeriod;

double glb_dEpsilon;;

int ( *dumpFunction )( double, int, double **, double**, double, void* );
void* glb_pDumpParam;
/*
 * These fuctions initiate and de initiate the Q- module
 */
void Init_Q_Module() {
  glb_nArrivalDistribution = 0;
  glb_nServiceDistribution = 0;
  glb_dSimulationTime = -1.0;
  glb_dDumpPeriod = -1.0;
  glb_dEvaluationPeriod = -1.0;

  glb_nClasses = glb_nResources = 0;
  glb_pClasses = NULL;
  glb_pResources = NULL;

  glb_nEventID = 1;
  glb_pEventList = NULL;

  glb_pZ = NULL;
  glb_pZOld = NULL;
  glb_pBlocking = NULL;
  glb_pN = NULL;
  glb_pNOld = NULL;
  glb_pRequired = NULL;
  glb_pServRate = NULL;
  glb_nAssoc = 0;
}

void Done_Q_Module() {
  unsigned int nIndex;
  RESOURCE* pRes;
  CLASS* pClass;
  
  /* Resources */
  for( nIndex = 0; nIndex < glb_nResources; nIndex++ ) {
    pRes = glb_pResources[ nIndex ];
    
    /* delete this resource */
    if( pRes != NULL ) {
      /* Delete Members */
      free( pRes->m_pClasses );
      free( pRes->m_pdBlockingRes );
      free( pRes->m_pdZ );
      free( pRes->m_pdN );
      free( pRes->m_pdNOld );
      free( pRes->m_pdServRate );
      free( pRes->m_pdRequired );
      free( pRes->m_pdZOld );
      free( pRes->m_pAssoc );
      
      free( pRes->m_pClasses );

      /* Delete the resource */
      free( pRes );
    }
  }
  
  /* Classes */
  for( nIndex = 0; nIndex < glb_nClasses; nIndex ++ ) {
    pClass = glb_pClasses[ nIndex ];

    /* delete this class*/
    if( pClass != NULL ) {
      /* Delete Members */
      free( pClass->m_pResources );
      free( pClass->m_pdBlockingRes );
      free( pClass->m_pdN );
      free( pClass->m_pdNOld );
      free( pClass->m_pdServRate );
      free( pClass->m_pdRequired );
      free( pClass->m_pdZ );
      free( pClass->m_pdZOld );
      free( pClass->m_pAssoc );

      free( pClass );
    }
  }

  /* Free the events */
  if( NULL != glb_pEventList ) {
    EVENT* pEvent, * pAuxEvent;
    for( nIndex = 0; nIndex < glb_nEvents; nIndex++ ) {
      pEvent = glb_pEventList[ nIndex ];
      while( pEvent != NULL ) {
	pAuxEvent = pEvent;
	pEvent = pEvent->m_pNext;
	free( pAuxEvent );
      }
    }

    free( glb_pEventList );
  }
 
  /* Free the globals */
  free( glb_pClasses );
  free( glb_pResources );
  free( glb_pZ );
  free( glb_pZOld );
  free( glb_pBlocking );
  free( glb_pN );
  free( glb_pNOld );
  free( glb_pRequired );
  free( glb_pServRate );
 
}

/*
 * Configuration parameters setting functions 
 * int Set_Config_i(int type, int value) {
 */

void Set_Config_i(int type, int value) {
  switch( type ) {

  case CONFIG_ARRIVAL_DISTRIBUTION: 
    glb_nArrivalDistribution = value;
    break;

  case CONFIG_SERVICE_DISTRIBUTION: 
    glb_nServiceDistribution = value;
    break;
  case CONFIG_NUMBER_OF_SIMULATIONS :
    /* Not in Z itteration */
    break;
  default:
    assert( 0 );
  }
}

void Set_Config_d(int type, double value) {
  switch( type ) {

  case CONFIG_SIMULATION_TIME: 
    glb_dSimulationTime = value;
    break;

  case CONFIG_DUMP_PERIOD: 
    glb_dDumpPeriod = value;
    break;

  case CONFIG_EVALUATION_PERIOD:
    glb_dEvaluationPeriod = value;
    break;
  case CONFIG_EPSILON_STOP:
    glb_dEpsilon = value;
    break;
  default:
    assert( 0 );
  }
}

/*
 * Adding / Deleting resources and classes
 */
int Add_Resource(unsigned int resourceID, double capacity) { /* add a single resource */
  RESOURCE* pNew;

  /* Allocate a new ressource */
  pNew = ( RESOURCE * ) malloc( sizeof( RESOURCE ));
  if( NULL == pNew )
    return ERR_ADD_ALLOC;

  /* init the members */
  pNew->m_nResourceID = resourceID;
  pNew->m_dUtilization = 0.0;
  pNew->m_dCapacity = capacity;
  pNew->m_nClasses = 0;
  pNew->m_pClasses = NULL;
  pNew->m_pdZ = NULL;
  pNew->m_pdZOld = NULL;
  pNew->m_pdBlockingRes = NULL;
  pNew->m_pAssoc = NULL;
  pNew->m_pdN = NULL;
  pNew->m_pdNOld = NULL;
  pNew->m_pdServRate = NULL;
  pNew->m_pdRequired = NULL;
  /* enlarge the ressources' array */
  glb_pResources = ( RESOURCE ** ) realloc( glb_pResources,  
					      ( glb_nResources + 1 ) * sizeof( RESOURCE * ));
  if( ESUCCESS != errno )
    return ERR_ADD_ALLOC;
  glb_pResources[ glb_nResources ] = pNew;
  glb_nResources++;

  return ERR_ADD_SUCCESS;
}


int Add_Class(unsigned int classID,  double dArrRate ){       /* adda single class */
  CLASS* pNew;

  /* Allocate a new ressource */
  pNew = ( CLASS * ) malloc( sizeof( CLASS ));
  if( NULL == pNew )
    return ERR_ADD_ALLOC;

  /* init the members */
  pNew->m_nClassID = classID;
  pNew->m_nResources = 0;
  pNew->m_pResources = 0;
  pNew->m_dBlocking = 0;
  pNew->m_pdZ = NULL;
  pNew->m_pdZOld = NULL;
  pNew->m_pdBlockingRes = NULL;
  pNew->m_pAssoc = NULL;
  pNew->m_dArrRate = dArrRate;
  pNew->m_dServRate = -1;
  pNew->m_bSameRate = TRUE; /* try to see if can minimize the computation */
  pNew->m_pdN = NULL;
  pNew->m_pdNOld = NULL;
  pNew->m_pdServRate = NULL;
  pNew->m_pdRequired = NULL;


  /* enlarge the classes' array */
  glb_pClasses = ( CLASS ** ) realloc( glb_pClasses,  
					      ( glb_nClasses + 1 ) * sizeof( CLASS * ));
  if( ESUCCESS != errno )
    return ERR_ADD_ALLOC;
  glb_pClasses[ glb_nClasses ] = pNew;
  glb_nClasses++;

  return ERR_ADD_SUCCESS;
}


int Add_Class_At_Resource(unsigned int classID, unsigned int resourceID,
			  double dMaxCapacity, double dServiceRate) {
  CLASS* pClass;
  RESOURCE* pResource;
  unsigned int nIndex, i;
  
  /* get the class and resource */
  pClass = getClassByID( classID );
  pResource = getResourceByID( resourceID );

  assert( NULL != pClass || NULL != pResource );
  
  /* reallocate */
  pResource->m_pClasses = ( struct tagCLASS ** ) realloc( pResource->m_pClasses, 
				    ( pResource->m_nClasses + 1 ) * sizeof( struct tagCLASS * ));
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  pClass->m_pResources = ( RESOURCE ** ) realloc( pClass->m_pResources, 
				    ( pClass->m_nResources + 1 ) * sizeof( RESOURCE * ));
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  /* also the pointers for Z and Blocking */
  pResource->m_pdZ = ( double ** ) realloc( pResource->m_pdZ, 
					    ( pResource->m_nClasses + 1 ) * sizeof( double * )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  pResource->m_pdZOld = ( double ** ) realloc( pResource->m_pdZOld, 
					    ( pResource->m_nClasses + 1 ) * sizeof( double * )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  pResource->m_pdBlockingRes = ( double ** ) realloc( pResource->m_pdBlockingRes, 
					    ( pResource->m_nClasses + 1 ) * sizeof( double * )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;
  
  pResource->m_pAssoc = ( unsigned int * ) realloc( pResource->m_pAssoc,
					   ( pResource->m_nClasses + 1 ) * sizeof( unsigned int )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  pResource->m_pdN = ( double ** ) realloc( pResource->m_pdN, 
					    ( pResource->m_nClasses + 1 ) * sizeof( double * )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  pResource->m_pdNOld = ( double ** ) realloc( pResource->m_pdNOld, 
					    ( pResource->m_nClasses + 1 ) * sizeof( double * )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  pResource->m_pdRequired = ( double ** ) realloc( pResource->m_pdRequired, 
					    ( pResource->m_nClasses + 1 ) * sizeof( double * )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  pResource->m_pdServRate = ( double ** ) realloc( pResource->m_pdServRate, 
					    ( pResource->m_nClasses + 1 ) * sizeof( double * )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  pClass->m_pdZ = ( double ** ) realloc( pClass->m_pdZ, 
					 ( pClass->m_nResources + 1 ) * sizeof( double * )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  pClass->m_pdZOld = ( double ** ) realloc( pClass->m_pdZOld, 
					 ( pClass->m_nResources + 1 ) * sizeof( double * )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  pClass->m_pdBlockingRes = ( double ** ) realloc( pClass->m_pdBlockingRes, 
						   ( pClass->m_nResources + 1 ) * sizeof( double * )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  pClass->m_pAssoc = ( unsigned int * ) realloc( pClass->m_pAssoc,
					( pClass->m_nResources + 1 ) * sizeof( unsigned int )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  pClass->m_pdN = ( double ** ) realloc( pClass->m_pdN, 
					 ( pClass->m_nResources + 1 ) * sizeof( double * )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  pClass->m_pdNOld = ( double ** ) realloc( pClass->m_pdNOld, 
					 ( pClass->m_nResources + 1 ) * sizeof( double * )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  pClass->m_pdRequired = ( double ** ) realloc( pClass->m_pdRequired, 
					 ( pClass->m_nResources + 1 ) * sizeof( double * )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  pClass->m_pdServRate = ( double ** ) realloc( pClass->m_pdServRate, 
					 ( pClass->m_nResources + 1 ) * sizeof( double * )) ;
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  /* Now reallocate the globals */
  glb_pZ = ( double * ) realloc( glb_pZ, ( glb_nAssoc + 1 ) * sizeof( double ));
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  glb_pZOld = ( double * ) realloc( glb_pZOld, ( glb_nAssoc + 1 ) * sizeof( double ));
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;
  
  glb_pBlocking = ( double * ) realloc( glb_pBlocking, ( glb_nAssoc + 1 ) * sizeof( double ));
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  glb_pN = ( double * ) realloc( glb_pN, ( glb_nAssoc + 1 ) * sizeof( double ));
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  glb_pNOld = ( double * ) realloc( glb_pNOld, ( glb_nAssoc + 1 ) * sizeof( double ));
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  glb_pRequired = ( double * ) realloc( glb_pRequired, ( glb_nAssoc + 1 ) * sizeof( double ));
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  glb_pServRate = ( double * ) realloc( glb_pServRate, ( glb_nAssoc + 1 ) * sizeof( double ));
  if( errno != ESUCCESS )
    return ERR_ADD_ALLOC;

  /* Set first the current association */
  pClass->m_pAssoc[ pClass->m_nResources ] = glb_nAssoc;
  pResource->m_pAssoc[ pResource->m_nClasses ] = glb_nAssoc;

  glb_nAssoc ++ ;

  glb_pZ[ glb_nAssoc - 1 ] = glb_pZOld[ glb_nAssoc - 1 ] = 0.0;
  glb_pN[ glb_nAssoc - 1 ] = glb_pNOld[ glb_nAssoc - 1 ] = 0.0;

  /* add the class tro ressource and the ressource to the class */
  pResource->m_pClasses[ pResource->m_nClasses ] = ( struct tagCLASS* ) pClass;
  pResource->m_nClasses++ ;	     

  pClass->m_pResources[ pClass->m_nResources ] = pResource;
  pClass->m_nResources++ ;

  /* Do the associations */
  for( i = 0; i < glb_nResources; i ++ ) {
    for( nIndex = 0; nIndex < glb_pResources[ i ]->m_nClasses; nIndex ++ ) {
      glb_pResources[ i ]->m_pdZ[ nIndex ] = &glb_pZ[ glb_pResources[ i ]->m_pAssoc[ nIndex ]];
      glb_pResources[ i ]->m_pdZOld[ nIndex ] = &glb_pZOld[ glb_pResources[ i ]->m_pAssoc[ nIndex ]];
      glb_pResources[ i ]->m_pdBlockingRes[ nIndex ] = &glb_pBlocking[ glb_pResources[ i ]->m_pAssoc[ nIndex ]];
      glb_pResources[ i ]->m_pdN[ nIndex ] = &glb_pN[ glb_pResources[ i ]->m_pAssoc[ nIndex ]];
      glb_pResources[ i ]->m_pdNOld[ nIndex ] = &glb_pNOld[ glb_pResources[ i ]->m_pAssoc[ nIndex ]];
      glb_pResources[ i ]->m_pdRequired[ nIndex ] = &glb_pRequired[ glb_pResources[ i ]->m_pAssoc[ nIndex ]];
      glb_pResources[ i ]->m_pdServRate[ nIndex ] = &glb_pServRate[ glb_pResources[ i ]->m_pAssoc[ nIndex ]];
    }
  }

  for( i = 0; i < glb_nClasses; i ++ ) {  
    for( nIndex = 0; nIndex < glb_pClasses[ i ]->m_nResources; nIndex ++ ) {
      glb_pClasses[ i ]->m_pdZ[ nIndex ] = &glb_pZ[ glb_pClasses[ i ]->m_pAssoc[ nIndex ]];
      glb_pClasses[ i ]->m_pdZOld[ nIndex ] = &glb_pZOld[ glb_pClasses[ i ]->m_pAssoc[ nIndex ]];
      glb_pClasses[ i ]->m_pdBlockingRes[ nIndex ] = &glb_pBlocking[ glb_pClasses[ i ]->m_pAssoc[ nIndex ]];
      glb_pClasses[ i ]->m_pdN[ nIndex ] = &glb_pN[ glb_pClasses[ i ]->m_pAssoc[ nIndex ]];
      glb_pClasses[ i ]->m_pdNOld[ nIndex ] = &glb_pNOld[ glb_pClasses[ i ]->m_pAssoc[ nIndex ]];
      glb_pClasses[ i ]->m_pdRequired[ nIndex ] = &glb_pRequired[ glb_pClasses[ i ]->m_pAssoc[ nIndex ]];
      glb_pClasses[ i ]->m_pdServRate[ nIndex ] = &glb_pServRate[ glb_pClasses[ i ]->m_pAssoc[ nIndex ]];
    }
  }
  
  /* Fill the structures with the input */
  glb_pRequired[ pClass->m_pAssoc[ pClass->m_nResources - 1 ] ] = dMaxCapacity;
  glb_pServRate[ pClass->m_pAssoc[ pClass->m_nResources - 1 ]] = dServiceRate;

  if( -1.0 == pClass->m_dServRate ) 
    pClass->m_dServRate = dServiceRate;
  else {
    if( dServiceRate != pClass->m_dServRate )
      pClass->m_bSameRate = FALSE;
  }

  return ERR_ADD_SUCCESS;
}

 
void Remove_Class(unsigned int classID) {
  unsigned int nIndex, mIndex;

  for( nIndex = 0; nIndex < glb_nClasses; nIndex ++ ) {

    if( glb_pClasses[ nIndex ] != NULL )
      if( glb_pClasses[ nIndex ]->m_nClassID == classID ) {

	/* Delete the class here */ 
	CLASS* pClass = glb_pClasses[ nIndex ];
	free( glb_pClasses[ nIndex ] );
	glb_pClasses[ nIndex ] = NULL;
	 
	/* delete the class everywhere */
	for( nIndex = 0; nIndex < glb_nResources; nIndex ++ ) {

	  if( glb_pResources[ nIndex ] != NULL ) {
	    for( mIndex = 0; mIndex < glb_pResources[ nIndex ]->m_nClasses; mIndex ++ ) {

	      if( ( CLASS * )glb_pResources[ nIndex ]->m_pClasses[ mIndex ] == pClass ) {
		glb_pResources[ nIndex ]->m_pClasses[ mIndex ] = NULL;
	      }
	    }
	  }
	}
      }
  }

  assert( FALSE );
}

void Remove_Resource(unsigned int resourceID){ 
  unsigned int nIndex, mIndex;

  for( nIndex = 0; nIndex < glb_nResources; nIndex ++ ) {

    if( glb_pResources[ nIndex ] != NULL )
      if( glb_pResources[ nIndex ]->m_nResourceID == resourceID ) {

	/* Delete the class here */ 
	RESOURCE* pResource = glb_pResources[ nIndex ];
	free( glb_pResources[ nIndex ] );
	glb_pResources[ nIndex ] = NULL;
	 
	/* delete the class everywhere */
	for( nIndex = 0; nIndex < glb_nClasses; nIndex ++ ) {

	  if( glb_pClasses[ nIndex ] != NULL ) {
	    for( mIndex = 0; mIndex < glb_pClasses[ nIndex ]->m_nResources; mIndex ++ ) {

	      if( glb_pClasses[ nIndex ]->m_pResources[ mIndex ] == pResource ) 
		glb_pClasses[ nIndex ]->m_pResources[ mIndex ] = NULL;
	    }
	  }
	}
      }
  }

  assert( FALSE );
}

/*
 * Callback registration
 */

int Add_Callback(int (*callback )( double, int, void * ),
		     double time, int nParam, void* pParam, unsigned int *pnId) {
  unsigned int nIndex;
  EVENT* pNew;
  /* If no event registered yet, create the list */

  if( NULL == glb_pEventList ) {
    /* Everything OK ? */
    if( glb_dSimulationTime <= 0 || glb_dEvaluationPeriod <= 0 ) 
      return ERR_CALLBACK_NOTINITIALIZED;

    /* Allocate it ! */
    nIndex = 1 + ( int ) ( glb_dSimulationTime / glb_dEvaluationPeriod );
    glb_nEvents = nIndex;
    glb_pEventList = malloc( nIndex * sizeof( EVENT* ));
    if( NULL == glb_pEventList ) 
      return ERR_CALLBACK_ALLOC;
    memset( glb_pEventList, 0, sizeof( EVENT* ) * nIndex );
  } 

  /* Another verification */
  if( time >= glb_dSimulationTime || time < 0.0 )
    return ERR_CALLBACK_BADTIME;

  /* add it */
  nIndex = ( unsigned int )( time / glb_dEvaluationPeriod );
  
  /* create the new event */
  pNew = ( EVENT * )malloc( sizeof( EVENT ));
  pNew->m_pCallback = callback;
  pNew->m_nParam = nParam;
  pNew->m_pParam = pParam;
  pNew->m_nID = glb_nEventID++;
  pNew->m_dTime = time;

  /* see what's next */
  pNew->m_pNext = glb_pEventList[ nIndex ];
  glb_pEventList[ nIndex ] = pNew;
}

void Remove_Callback(unsigned int ID) {
  unsigned int nIndex;
  EVENT * pFront, * pBack;

  for( nIndex = 0; nIndex < glb_nEvents; nIndex ++ ) {
    /* find in current bucket */
    pBack = NULL;
    pFront = glb_pEventList[ nIndex ];
    
    /* Find it.. */
    while( pFront != NULL ) {
      if( pFront->m_nID == ID ) {
	/* restore the links */
	if( pBack != NULL )
	  pBack->m_pNext = pFront->m_pNext;
	else
	  glb_pEventList[ nIndex ] = pFront->m_pNext;

	/* delete it ! */
	free( pFront );
	return;
      }
    }
  }

  assert( FALSE );
}

 
void Set_Arrival_Rate(unsigned int classID, double dNewRate, double *dOldRate) {
  /* Get the class */
  CLASS *pClass;
  pClass = getClassByID( classID );

  /* set the rate */
  if( NULL != dOldRate )
    *dOldRate = pClass->m_dArrRate;
  pClass->m_dArrRate = dNewRate;
}

void Set_Service_Rate( unsigned int classID, unsigned int nResourceID, 
		       double dNewRate, double *dOldRate) {
  /* Get the class */
  CLASS *pClass;
  unsigned int nIndex;
  pClass = getClassByID( classID );

  for( nIndex = 0; nIndex < pClass->m_nResources; nIndex ++ ) {
    if( pClass->m_pResources[ nIndex ]->m_nResourceID == nResourceID ) {
      if( NULL != dOldRate )
	*dOldRate = *pClass->m_pdServRate[ nIndex ];
      *pClass->m_pdServRate[ nIndex ] = dNewRate;
      return;
    }
  }

  assert( FALSE );
}



/*
 * Dump management
 */
void Register_Dump_Function( int (*callback)( double, int, double**, double**, double, void* ),
			     void* pDumpParam) {
  dumpFunction = callback;
  glb_pDumpParam = pDumpParam;
}

/*
 * Fetching measures
 */ 
/* Old Names
#define M_BLOCKING_PROB_1         1
#define M_BLOCKING_PROB_2         2
#define M_NUMBER_OF_OCCUPANTS     3
#define M_UTILIZATION             4
*/
#define METRIC_BLOCKING_PROB_CLASS     1
#define METRIC_BLOCKING_PROB           2
#define METRIC_NUMBER_CLASS			   3
#define METRIC_UTILIZATION             4

int Fetch_Crt_Basic_Metrics( int MeasureID,unsigned int ClassID,
			      unsigned int ResourceID, double *value ) {
  unsigned int nIndex ;

  switch( MeasureID ) {
  case METRIC_BLOCKING_PROB_CLASS :
    *value = glb_pClasses[ ClassID ]->m_dBlocking;
    break;
  case METRIC_BLOCKING_PROB :
    for( nIndex = 0; nIndex < glb_pClasses[ ClassID ]->m_nResources; nIndex ++ ){ 
      if( glb_pClasses[ ClassID ]->m_pResources[ nIndex ]->m_nResourceID == ResourceID )
	*value = *glb_pClasses[ ClassID ]->m_pdBlockingRes[ ResourceID ];
    };
    break;
  case METRIC_NUMBER_CLASS :
    *value = glb_pClasses[ ClassID ]->m_dN;
    break;   
  case METRIC_UTILIZATION :
    *value = glb_pResources[ ResourceID ]->m_dUtilization;
    break;
  default :
    return ERR_FETCH_NOMETRIC;
  }

  return ERR_FETCH_SUCCESS;
}

int Fetch_Avg_Basic_Metrics( int MeasureID,unsigned int ClassID,
			      unsigned int ResourceID, double *value ) {
  return Fetch_Crt_Basic_Metrics(MeasureID, ClassID, ResourceID, value );
}


/*
 * This will compute the utilization 
 */
void Recompute_Resource_Utilization( ) {
  ComputeUtilization( glb_pClasses, glb_nClasses, 
		      glb_pResources, glb_nResources );
}

extern unsigned int glb_nClasses;
extern CLASS ** glb_pClasses;

extern unsigned int glb_nResources;
extern RESOURCE ** glb_pResources;
extern unsigned int glb_nEvents;

/*
 * The real meat
 */
void Evaluate( ) {
  DoComputation( glb_pClasses, glb_nClasses, 
		 glb_pResources, glb_nResources,
		 glb_dSimulationTime, glb_dEvaluationPeriod,
		 glb_pEventList, glb_dEpsilon, glb_dDumpPeriod );
}

/*
 * This displays a message for an error id
 * --- NOTHING NOW ---
 */
void Print_Mcmr_Error(int errcode) {
  printf( " Some error here .. \n " );
}




