/*
 * zit.h
 *
 * This file contains the central function from z-iteration module
 */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "mcmr.h"
#include "structures.h"

/*
 * This function will compute the utilization for all the links
 *
 * double dLinkUtilization[] - the utilization for each link
 * unsigned int nLinks - the number of links
 * TRAFFIC_CLASSES *pClasses - where to sera
 */
void ComputeUtilization( CLASS** pClasses, unsigned int nClasses,  
			 RESOURCE** pResources, unsigned int nResources ) {

  unsigned int nIndex, nInnerIndex;

  /* Reset the utilizations */
  for( nIndex = 0; nIndex < nResources; nIndex ++ ) 
    pResources[ nIndex ]->m_dUtilization = 0.0;

  for( nIndex = 0; nIndex < nClasses; nIndex ++ ) {
    /* For each class increment the resources' utilization */
    for( nInnerIndex = 0; nInnerIndex < pClasses[ nIndex ]->m_nResources; nInnerIndex ++ ) {
      pClasses[ nIndex ]->m_pResources[ nInnerIndex ]->m_dUtilization +=
	( *pClasses[ nIndex ]->m_pdN[ nInnerIndex ] ) * ( *pClasses[ nIndex ]->m_pdRequired[ nInnerIndex ] );
    }
  }

  /* Reset the utilizations */
  for( nIndex = 0; nIndex < nResources; nIndex ++ ) 
    pResources[ nIndex ]->m_dUtilization = pResources[ nIndex ]->m_dUtilization / 
      pResources[ nIndex ]->m_dCapacity;

}

/*
 * This computes the blocking probability for each class which belongs to 
 * a specified resource
 *
 * Input :
 *   RESOURCE* pRes - the resource for which the blocking probabilities have to be 
 *  computed
 *
 */
void ComputeBlocking( RESOURCE* pRes ) { 
  int nIndex, nIndexInner;
  unsigned int nBandWidth;
  double* Q;
  double dSum;

  nBandWidth = ( unsigned int )pRes->m_dCapacity;
  
  /* Allocate the array of Q's */
  Q = malloc( sizeof( double ) * ( nBandWidth + 1 ));
  if( NULL == Q ) {
    fprintf( stderr," Insufficient Memory! \n " );
    exit( -1 );
  }
  
  Q[ 0 ] = 1.0;
  dSum = Q[ 0 ];

  /* Compute all the other Q-s */
  for( nIndex = 1; nIndex <= nBandWidth; nIndex ++ ) {
    Q[ nIndex ] = 0;
    for( nIndexInner = 0; nIndexInner < pRes->m_nClasses; nIndexInner ++ )
      if( NULL != pRes->m_pClasses[ nIndexInner ] ) {
	if( nIndex >= ( int )( *pRes->m_pdRequired[ nIndexInner ] ))
	  Q[ nIndex ] += ( *pRes->m_pdRequired[ nIndexInner ] )* 
	    ( *(pRes->m_pdZ[ nIndexInner ]) ) *
	    Q[ nIndex - ( int )( *pRes->m_pdRequired[ nIndexInner ] ) ];
      }
    Q[ nIndex ] /= ( double )nIndex ;  
    dSum += Q[ nIndex ];
  }
  
  /* Compute the blocking probalility */
  for( nIndex = 0; nIndex < pRes->m_nClasses; nIndex ++ ) {
      if( NULL != pRes->m_pClasses[ nIndex ] ) {
	*(pRes->m_pdBlockingRes[ nIndex ]) = 0.0;
	if( dSum > 0 ) {
	  for(  nIndexInner = 0;
		nIndexInner <  nBandWidth - 
		  ( int )( *pRes->m_pdRequired[ nIndex ] )+ 1;
		nIndexInner ++ ) 
	    *(pRes->m_pdBlockingRes[ nIndex ]) += Q[ nIndexInner ];
	  *(pRes->m_pdBlockingRes[ nIndex ]) = ( dSum - *(pRes->m_pdBlockingRes[ nIndex ])) / dSum;
	}
      }
  }
  
  free( Q );
}
  
/*
 * This does the Z iteration for a specific resource to determine the 
 * blocking probability.
 *
 * Input :
 *    RESOURCE* pRes - the resource for which to do the iteration
 *    double dEpsilon - an epsilon where to stop;
 */
void DoIteration( RESOURCE* pRes, double dEpsilon ) {
  int nIndex, nInnerIndex;
  int bRecompute;

  do {
    for( nIndex = 0; nIndex < pRes->m_nClasses; nIndex ++ )  
      if( pRes->m_pClasses[ nIndex ] != NULL )
	*(pRes->m_pdZ[ nIndex ]) = *(pRes->m_pdZOld[ nIndex ]);

    ComputeBlocking( pRes );

    for( nIndex = 0; nIndex < pRes->m_nClasses; nIndex ++ ) 
      if( pRes->m_pClasses[ nIndex ] != NULL )
	*(pRes->m_pdZOld[ nIndex ]) = 
	  ( 1 / ( 1 - *pRes->m_pdBlockingRes[ nIndex ] )) * 
	  ( *pRes->m_pdNOld[ nIndex ] ) ;
    
    bRecompute = FALSE;
    for( nIndex = 0; nIndex < pRes->m_nClasses && !bRecompute ; nIndex ++ ) {
	if( pRes->m_pClasses[ nIndex ] != NULL ) {
	  double dDif = *(pRes->m_pdZ[ nIndex ]) - *(pRes->m_pdZOld[ nIndex ]);
	  if( dDif < 0 )
	    dDif = -dDif;
	  if( dDif > dEpsilon ) 
	    bRecompute = TRUE;
	}
    }
  } while( bRecompute );
}
	 
/* The computation itself ... the mother of all
 *
 * Input :
 *   CLASS* classes[] - the classes 
 *   unsigned int nClasses - the # of classes
 *   RESOURCE* resources[] - the resources
 *   unsigned int nResources - the # of resources
 *   double dInterval - how long to evaluate
 *   double dDelta - the step
 *   EVENT* events[] - the events to take care of 
 *   double dEpsilon - the stop limit epsilon */
void DoComputation( CLASS* classes[], unsigned int nClasses,
		    RESOURCE* resources[], unsigned int nResources,
		    double dInterval, double dDelta,
		    EVENT* events[],
		    double dEpsilon, double dDump ) {
  int nIndex, nInnerIndex, nStep, nDoN ;
  /* indexes arround here */
  double dCrtTime, dCrtDump;
  /* the current simulation time */
  double dFactorLeft, dFactorRight;
  /* some auxiliary factors */
  double dL, dM;
  /* this will hold current lambda / miu */
  EVENT *pEvent;

  dCrtDump = 0;
  /* step through the time .. */
  for( dCrtTime = 0, nStep = 0; dCrtTime < dInterval; dCrtTime += dDelta, nStep++ ) {
    
    /* update the dump counter */
    dCrtDump += dDelta;

    /* check for some events .. and do them */
    if( events != NULL ) {
      pEvent = events[ nStep ];
      while( NULL != pEvent ) {
	pEvent->m_pCallback( dCrtTime, pEvent->m_nParam, pEvent->m_pParam );
	pEvent = pEvent->m_pNext;
      }
    }

    /* update the N */
    for( nIndex = 0; nIndex < nClasses; nIndex ++ ) 
      for( nInnerIndex = 0; nInnerIndex < classes[ nIndex ]->m_nResources; nInnerIndex ++ )
	( *classes[ nIndex ]->m_pdNOld[ nInnerIndex ] )= ( *classes[ nIndex ]->m_pdN[ nInnerIndex ] );
    
    /* Do the iteration for each resource */  
    for( nIndex = 0; nIndex < nResources; nIndex ++ )
      DoIteration( resources[ nIndex ], dEpsilon );
    
    for( nIndex = 0; nIndex < nClasses; nIndex ++ ){
      if( NULL != classes[ nIndex ] ) {
	
	/* Get it's miu and lamda */
	dL = classes[ nIndex ]->m_dArrRate;

	/*	dM = classes[ nIndex ]->m_dServRate; */

	if( TRUE /*!rateFunction( dCrtTime, classes[ nIndex ]->m_nClassID, &dL, &dM ) */ ) {
	  
	  /* Sum the blocking */
	  classes[ nIndex ]->m_dTotalAdmission = 1.0; 
	  for( nInnerIndex = 0; nInnerIndex < classes[ nIndex ]->m_nResources ; nInnerIndex ++ )
	    if( classes[ nIndex ]->m_pResources[ nInnerIndex ] != NULL )
	      classes[ nIndex ]->m_dTotalAdmission *= 
		1 - *(classes[ nIndex ]->m_pdBlockingRes[ nInnerIndex ]);
	  
	  /* same service rate on all links ? */
	  if( classes[ nIndex ]->m_bSameRate ) 
	    nDoN = 1 > classes[ nIndex ]->m_nResources? 0 : 1;
	  else
	    nDoN = classes[ nIndex ]->m_nResources;

	  for( nInnerIndex = 0; nInnerIndex < nDoN; nInnerIndex ++  ) {
	    if( classes[ nIndex ]->m_pResources[ nInnerIndex ] != NULL ) {
	      dM = *classes[ nIndex ]->m_pdServRate[ nInnerIndex ];
	      
	      /* compute some auxiliary factors */
	      dFactorLeft =  1 - dDelta * dM ;
	      dFactorRight = dL*classes[ nIndex ]->m_dTotalAdmission;
	    
	      /* Now get the N */
	      *classes[ nIndex ]->m_pdN[ nInnerIndex ] = 
		dFactorLeft * ( *classes[ nIndex ]->m_pdNOld[ nInnerIndex ] ) + dDelta * dFactorRight;
	  
	      /* Here can come this .. experience says */
	      if( *classes[ nIndex ]->m_pdN[ nInnerIndex ] < 0 ) {
		fprintf( stderr, " Error: delta too large..\n" );
		*classes[ nIndex ]->m_pdN[ nInnerIndex ] = 0;
	      }
	    }
	    else {
	      if( nDoN < classes[ nIndex ]->m_nResources &&
		  classes[ nIndex ]->m_bSameRate )
		nDoN ++;
	    }
	  }

	  /* If i done some speed - ups, update all Ns */
	  if( nDoN != classes[ nIndex ]->m_nResources ) {
	    for( nInnerIndex = nDoN; nInnerIndex < classes[ nIndex ]->m_nResources; nInnerIndex ++  ) {
	      ( *classes[ nIndex ]->m_pdN[ nInnerIndex ] )= 
		( *classes[ nIndex ]->m_pdN[ nDoN - 1 ] );
	    }
	  }
	  
	  /* Report the data */
	  if( dCrtDump >= dDump ) {
	    if( NULL != dumpFunction ) 
	      dumpFunction( dCrtTime, classes[ nIndex ]->m_nClassID, 
			    classes[ nIndex ]->m_pdN,
			    classes[ nIndex ]->m_pdBlockingRes,
			    classes[ nIndex ]->m_dTotalAdmission,
			    glb_pDumpParam );
	  }
	}  
      }
    }
    
    /* update the dump counter */
    if( dCrtDump >= dDump )
      dCrtDump = 0;
  }
}
